"""
ImageNode is a Python Class to govern the data and metadata of image data (OpenCV, PIL). 

This node builds upon the OpenCV and PIL API for the image definitions.
It directly inherits from Node.
Be sure to check the properties defined in the above classes to initialise the Node.
"""
#IMPORT PACKAGES
from ast import Raise
from distutils import extension
import xml.etree.ElementTree as ET
import cv2
import PIL
from rdflib import Graph, URIRef
import numpy as np
import os
import open3d as o3d
import math
import uuid
from scipy.spatial.transform import Rotation as R


#IMPORT MODULES
from geomapi.nodes import Node
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt


class ImageNode(Node):
    # class attributes
    
    def __init__(self,  graph : Graph = None, 
                        graphPath:str=None,
                        subject : URIRef = None,
                        path : str=None, 
                        xmpPath: str = None,
                        xmlPath: str = None,
                        getResource : bool = False,
                        getMetaData : bool = True,
                        **kwargs): 
        """Creates a Node from one or more of the following inputs. 
        By default, no data is imported in the Node to speed up processing.
        If you also want the data, call node.get_resource() or set getResource() to True.\n

        Args: \n
            1. graph (RDFlib Graph) : Graph with a single subject (if multiple subjects are present, only the first will be used to initialise the MeshNode)\n
            2. graphPath (str):  Graph file path with a single subject (if multiple subjects are present, only the first will be used to initialise the MeshNode)\n
            3. path (str) : path to image file (Note that this node will also contain the data) \n
            4. resource (ndarray, PIL Image,Open3D) : OpenCV, PIL (Note that this node will also contain the data)\n
            5. xmlPath (str) : Xml file path from Agisoft Metashape\n
            6. xmpPath (str) :  xmp file path from RealityCapture
            \n
            - getResource (bool, optional= False): If True, the node will search for its physical resource on drive \n
            - getMetaData (bool, optional= True): If True, the node will attempt to extract metadata from the resource if present \n

        Returns:
            An ImageNode with metadata 
        """  
        #private attributes 
        self._xmlPath=None
        self._xmpPath=None
        self._orientedBoundingBox=None
        self.imageWidth = None # (int) number of pixels
        self.imageHeight = None  # (int) number of pixels
        self.focalLength35mm = None # (Float) focal length in mm     

        super().__init__(   graph= graph,
                            graphPath= graphPath,
                            subject= subject,
                            path=path,
                            **kwargs)   

        #instance variables
        self.xmlPath=xmlPath
        self.xmpPath=xmpPath

        #initialisation functionality
        if getMetaData:
            if self.get_metadata_from_xmp_path():
                pass
            elif self.get_metadata_from_xml_path():
                pass

        if getResource:
            self.get_resource()

        if getMetaData:
            self.get_metadata_from_exif_data()
            self.get_metadata_from_resource()

#---------------------PROPERTIES----------------------------

    #---------------------xmlPath----------------------------
    @property
    def xmlPath(self): 
        """Get the xmlPath (str) of the node."""
        return self._xmlPath

    @xmlPath.setter
    def xmlPath(self,value):
        if value is None:
            return None
        elif (type(value) == str and value.endswith('xml') ):
            self._xmlPath=value
        else:
            raise ValueError('self.xmlPath has invalid type, path or extension.')    

  #---------------------xmpPath----------------------------
    @property
    def xmpPath(self): 
        """Get the xmpPath (str) of the node."""
        return self._xmpPath

    @xmpPath.setter
    def xmpPath(self,value):
        if value is None:
            return None
        elif (type(value) == str and value.endswith('xmp') ):
            self._xmpPath=value
        else:
            raise ValueError('self.xmpPath has invalid type, path or extension.')    

#---------------------orientedBoundingBox----------------------------
    @property
    def orientedBoundingBox(self): 
        """Get the orientedBoundingBox of the Node from various inputs. \n

        Args:
            1. Open3D.geometry.OrientedBoundingBox \n
            2. Open3D geometry\n

        Returns:
            orientedBoundingBox (o3d.geometry.OrientedBoundingBox) 
        """
        return self._orientedBoundingBox

    @orientedBoundingBox.setter
    def orientedBoundingBox(self,value):
        if value is None:
            return None
        if 'orientedBoundingBox' in str(type(value)):
            self._orientedBoundingBox=value
        else:    
            try: #geometry
                self._orientedBoundingBox=value.get_oriented_bounding_box()
            except:
                raise ValueError('Input must be orientedBoundingBox (o3d.geometry.OrientedBoundingBox) or an Open3D Geometry.')

#---------------------METHODS----------------------------
   
    def set_resource(self,value):
        """Set the resource of the Node from various inputs.\n

        Args:
            1. np.ndarray (OpenCV) \n
            2. PIL Image\n
            3. Open3D Image\n

        Raises:
            ValueError: Resource must be np.ndarray (OpenCV), PIL Image or Open3D Image.
        """

        if type(value) is np.ndarray : #OpenCV
            self._resource = value
        elif 'Image' in str(type(value)): # PIL
            self._resource=  cv2.cvtColor(np.array(value), cv2.COLOR_RGB2BGR)
        else:
            raise ValueError('Resource must be np.ndarray (OpenCV) or PIL Image')

    def get_resource(self)->np.ndarray: 
        """Returns the resource (image) in the node. 
        If none is present, it will search for the data on drive from the following inputs. \n

        Args:
            1. self.path\n
            2. self.graphPath\n
            3. self.name or self.subject

        Returns:
            np.ndarray or None
        """
        if self.resource is not None :
            pass
        elif self.get_path():
            self._resource   = cv2.imread(self.path)
        return self._resource  

    def get_path(self) -> str:
        """Returns the full path of the resource.
        If none is present, it will search for the data on drive from the following inputs.\n

        Args:
            1. self.graphPath \n
            2. self.name \n
            3. self.subject\n

        Returns:
            resource path (str)
        """      
        if self._path and os.path.exists(self._path):
            return self._path
        nodeExtensions=ut.get_node_resource_extensions(str(type(self)))
        if self._graphPath and (self._name or self._subject):
            folder=ut.get_folder_path(self._graphPath)
            allSessionFilePaths=ut.get_list_of_files(folder) 
            for path in allSessionFilePaths:
                if ut.get_extension(path) in nodeExtensions:
                    if self.get_name() in path or self.get_subject() in path :
                        self._path = path    
                        return self._path
            if self._name:
                self._path=os.path.join(folder,self._name+nodeExtensions[0])
            else:
                self._path=os.path.join(folder,self._subject+nodeExtensions[0])
            return self._path
        elif self._xmpPath and os.path.exists(self._xmpPath):
            folder=ut.get_folder_path(self._xmpPath)
            allSessionFilePaths=ut.get_list_of_files(folder) 
            for path in allSessionFilePaths:
                if ut.get_extension(path) in nodeExtensions:
                    if ut.get_filename(self._xmpPath) in path :
                        self.name=ut.get_filename(self._xmpPath)
                        self.subject=self._name
                        self.path = path    
                        return self._path
        elif self._xmlPath and os.path.exists(self._xmlPath):
            folder=ut.get_folder_path(self._xmlPath)
            allSessionFilePaths=ut.get_list_of_files(folder) 
            for path in allSessionFilePaths:
                if ut.get_extension(path) in nodeExtensions:
                    if self.get_name() in path or self.get_subject() in path :
                        self._path = path    
                        return self._path
        else:
            # print("No file containing this object's name and extension is found in the graphPath folder")
            return None

    def get_xmp_path(self)->str: 
        """Returns the xmpPath in the node. 
        If none is present, it will search for the data on drive from the following inputs.\n

        Args:
            1. self.graphPath \n
            2. self.name \n
            3. self.subject\n

        Returns:
            str or None
        """
        if self._xmpPath and os.path.exists(self._xmpPath):
            return self._xmpPath            
        elif self._graphPath and (self._name or self._subject):
            folder=ut.get_folder_path(self._graphPath)
            allSessionFilePaths=ut.get_list_of_files(folder) 
            for path in allSessionFilePaths:
                if ut.get_extension(path).endswith('xmp'):
                    if self.get_name() in path or self.get_subject() in path :
                        self._xmpPath = path    
                        return self._xmpPath
        else:
            return None

    def save_resource(self, directory:str=None,extension :str = '.png') ->bool:
        """Export the resource of the Node.\n

        Args:
            1. directory (str, optional): directory folder to store the data.\n
            2. extension (str, optional): file extension. Defaults to '.png'.\n

        Raises:
            ValueError: Unsuitable extension. Please check permitted extension types in utils._init_.\n

        Returns:
            bool: return True if export was succesful
        """     
        #check path
        if self.resource is None:
            return False
        
        #validate extension
        if extension not in ut.IMG_EXTENSION:
            raise ValueError('Invalid extension')

        # check if already exists
        if directory and os.path.exists(os.path.join(directory,self.get_name() + extension)):
            self.path=os.path.join(directory,self.get_name() + extension)
            return True
        elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.IMG_EXTENSION:
            return True    
          
        #get directory
        if (directory):
            pass    
        elif self.path is not None and os.path.exists(self.path):    
            directory=ut.get_folder(self.path)            
        elif(self.graphPath): 
            dir=ut.get_folder(self.graphPath)
            directory=os.path.join(dir,'IMG')   
        else:
            directory=os.path.join(os.getcwd(),'IMG')
        # create directory if not present
        if not os.path.exists(directory):                        
            os.mkdir(directory)       

        self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)

        #write files
        if cv2.imwrite(self.path, self.resource):
            return True
        return False
       
    def get_oriented_bounding_box(self)->o3d.geometry.OrientedBoundingBox:
        """Gets the Open3D OrientedBoundingBox of the node from the conical mesh representation based on the 
        cartesianTransform, the focal length at 35mm and a viewing range. \n

        Returns:
            o3d.geometry.orientedBoundingBox
        """    
        if self._orientedBoundingBox is not None:
            pass
        elif self._cartesianTransform is not None:
            mesh=self.get_mesh_geometry()
            self._orientedBoundingBox=mesh.get_oriented_bounding_box()            
        else:
            return None
        return self._orientedBoundingBox

    def get_metadata_from_resource(self) ->bool:
        """Returns the metadata from a resource. \n

        Features:
            1. imageHeight\n
            2. imageWidth\n

        Returns:
            bool: True if exif data is successfully parsed
        """
        if self._resource is None:
            return False   
              
        try:
            if getattr(self,'imageHeight',None) is None:
                self.imageHeight=self.resource.shape[0]
            if getattr(self,'imageWidth',None) is None:
                self.imageWidth=self.resource.shape[1]
            return True
        except:
            raise ValueError('Metadata extraction from resource failed')
    
    # def get_oriented_bounding_box(self)->o3d.geometry.OrientedBoundingBox:
    #     """Gets the Open3D geometry from cartesianTransform

    #     Returns:
    #         o3d.geometry.orientedBoundingBox
    #     """
    #     if getattr(self,'orientedBoundingBox',None) is None:                
    #         if getattr(self,'cartesianTransform',None) is not None:
    #             box=o3d.geometry.create_mesh_box(width=1.0, height=1.0, depth=1.0)
    #             self.orientedBoundingBox= box.transform(self.cartesianTransform)
    #         else:
    #             return None
    #     return self.orientedBoundingBox

    def get_mesh_geometry(self, depth:float=10, focalLength35mm:float=24)->o3d.geometry.TriangleMesh:
        """Generate a concical mesh representation using the Image's cartesianTransform and focalLength35mm.\n
    
        .. image:: ../../../docs/pics/virtual_image2.PNG

        Args:
            1. depth (float, optional): Viewing depth of the image. Defaults to 10m.\n
            2. focalLength35mm (float,optional): standardised focal length on 35mm film (w=36mm, h = 24mm)\n

        Returns:
            o3d.geometry.TriangleMesh 
        """
        if self.cartesianTransform is not None:
            radius=35/(focalLength35mm*2)*depth        
            mesh= o3d.geometry.TriangleMesh.create_cone(radius=radius, height=depth, resolution=20, split=1)
            rotation=gt.get_rotation_matrix(self.cartesianTransform)
            r=R.from_matrix(rotation)
            rz=R.from_euler('xyz' ,[180, 0, 0], degrees=True)
            t=gt.get_translation(self.cartesianTransform)
            mesh=mesh.translate(t)
            r=rz*r
            # t2=r.as_matrix() * np.array([[1],[0],[0]]) *depth
            A = np.dot( r.as_matrix(),np.array([0,0,-1]) )*depth
            mesh=mesh.translate(A)
            rot=r.as_matrix()
            mesh=mesh.rotate(rot)
            return mesh
        else:
            return None

    def get_virtual_image(self, geometries: o3d.geometry, downsampling:int=2)-> o3d.cpu.pybind.geometry.Image:
        """Generates a virtual image of a set of geometries given the ImageNode's pose and piholeModel.

        .. image:: ../../../docs/pics/rendering3.PNG


        Args:
            1. geometries (o3d.geometry): geometries to include in the scene of the virtual image.\n
            2. downsampling (int, optional): pixel downsampling of the image both in height and width (each step reduces the density by factor 4). Defaults to 2.

        Returns:
            o3d.cpu.pybind.geometry.Image or None
        """
        pinholeCamera=self.get_pinhole_camera_parameters(downsampling)
        if pinholeCamera is not None:
            return gt.generate_virtual_image(geometries,pinholeCamera)
        else:
            return None

    def get_pinhole_camera_parameters(self, downsampling:int=2) -> o3d.camera.PinholeCameraParameters():
        """Returns the intrinsic and extrinsix camera parameters based on the following attributes.

        .. image:: ../../../docs/pics/pinholemodel1.PNG

        Args:
            1. self.imageWidth: width of the image in pixels (u) \n
            2. self.imageHeight: height of the image in pixels (v) \n
            3. self.focalLength35mm: focal length with a standardised Field-of-View.\n 
            4. self.cartesianTransform: the inverted transform equals the external camera pose.\n
            2. downsampling (int, optional): pixel downsampling of the image both in height and width (each step reduces the density by factor 4). Defaults to 2.

        Returns:
            o3d.camera.PinholeCameraParameters()
        """
        param=o3d.camera.PinholeCameraParameters()
        if getattr(self,'cartesianTransform',None) is not None:
            param.extrinsic=np.linalg.inv(self.cartesianTransform)
            param.intrinsic=self.get_intrinsic_camera_parameters(downsampling)
            self.pinholeCamera=param
            return self.pinholeCamera
        else:
            return None

    def get_intrinsic_camera_parameters(self, downsampling:int=2) -> o3d.camera.PinholeCameraIntrinsic():
        """Returns the intrinsic camera parameters based on the following attributes.
        
        Args:
            1. self.imageWidth: width of the image in pixels (u). Defaults to 640p \n
            2. self.imageHeight: height of the image in pixels (v). Defaults to 480p  \n
            3. self.focalLength35mm: focal length with a standardised Field-of-View. Defaults to 25mm \n 
            4. self.PrincipalPointU: cx \n
            4. self.PrincipalPointV: cy \n

        Returns:
            o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)
        """
        width=640
        height=480
        f=25

        if getattr(self,'imageWidth',None) is not None:
            width=int(self.imageWidth/downsampling)
        if getattr(self,'imageHeight',None) is not None:
            height=int(self.imageHeight/downsampling)
        if getattr(self,'focalLength35mm',None) is not None:
            f=self.focalLength35mm

        pixX=width/36 #these are standard 35mm film properties
        pixY=height/24 #these are standard 35mm film properties
        fx=pixX*f
        fy=pixY*f

        if (getattr(self,'principalPointU',None) is not None and
            getattr(self,'principalPointV',None) is not None ):
            cx=width/2-0.5+self.principalPointU
            cy=height/2-0.5+self.principalPointV
        else:
            cx=width/2-0.5
            cy=height/2-0.5
        return o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)

    def get_metadata_from_exif_data(self) -> bool:
        """Returns the metadata from a resource. \n

        Features:
            1. GPSInfo (geospatialTransform (np.array(3,1))
            2. coordinateSystem (str) \n
            2. DateTime ('%Y-%m-%dT%H:%M:%S')\n
            3. XResolution (int)\n
            4. YResolution (int)\n
            5. ResolutionUnit (int)\n
            6. ExifImageWidth (int)\n
            7. ExifImageHeight (int)\n

        Returns:
            bool: True if meta data is successfully parsed
        """
        if  self.get_path() is None or not os.path.exists(self.get_path()) :
            return False
        
        if getattr(self,'timestamp',None) is None :
            self.timestamp=ut.get_timestamp(self.path)
        
        if getattr(self,'name',None) is None:
            self.name=ut.get_filename(self.path)

        if (getattr(self,'imageWidth',None) is not None and
            getattr(self,'imageHeight',None) is not None and
            getattr(self,'geospatialTransform',None) is not None):
            return True

        # pix = PIL.Image.open(self.path) 
        with PIL.Image.open(self.path) as pix:
            exifData=ut.get_exif_data(pix)

        if exifData is not None:
            self.timestamp=ut.get_if_exist(exifData, "DateTime")
            self.resolutionUnit=ut.get_if_exist(exifData,"ResolutionUnit")
            self.imageWidth=ut.get_if_exist(exifData,"ExifImageWidth")
            self.imageHeight=ut.get_if_exist(exifData,"ExifImageHeight")
            
            if 'GPSInfo' in exifData:
                gps_info = exifData["GPSInfo"]
                if gps_info is not None:
                    # self.GlobalPose=GlobalPose # (structure) SphericalTranslation(lat,long,alt), Quaternion(qw,qx,qy,qz)
                    latitude=ut.get_if_exist(gps_info, "GPSLatitude")
                    latReference=ut.get_if_exist(gps_info, "GPSLatitudeRef")
                    newLatitude=ut.filter_exif_gps_data(latitude,latReference)
                    longitude= ut.get_if_exist(gps_info, "GPSLongitude")
                    longReference=ut.get_if_exist(gps_info, "GPSLongitudeRef")
                    newLongitude=ut.filter_exif_gps_data(longitude,longReference)
                    self.geospatialTransform=[  newLatitude, 
                                                newLongitude,
                                                ut.get_if_exist(gps_info, "GPSAltitude")]
                    self.coordinateSystem='geospatial-wgs84'
            
            return True
        else:
            return False
    
    def get_metadata_from_xmp_path(self)->bool:
        """Read Metadata from .xmp file generated by https://www.capturingreality.com/.

        Features:
            1. geospatialTransform (np.array(3x1))\n
            2. coordinateSystem (str)\n
            3. focalLength35mm (float)\n
            4. principalPointU (float)\n
            5. principalPointV (float)\n
            6. cartesianTransform (np.array(4x4))\n

        Returns:
            bool: True if metadata is sucesfully parsed
        """
        if self.xmpPath is None or not os.path.exists(self.xmpPath):
            return False

        if (getattr(self,'principalPointU',None) is not None and
            getattr(self,'principalPointV',None) is not None and
            getattr(self,'distortionCoeficients',None) is not None and
            getattr(self,'geospatialTransform',None) is not None ):         
            return True
        
        mytree = ET.parse(self.xmpPath)
        root = mytree.getroot()                       
        
        self.timestamp=ut.get_timestamp(self.xmpPath)
        self.name=ut.get_filename(self.xmpPath)
        self.subject=self.name
        for child in root.iter('{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description'):

            #Attributes
            for attribute in child.attrib:
                if ('latitude' in attribute and
                    'longitude'in attribute and
                    'altitude' in attribute):
                    lat=ut.xcr_to_lat(child.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}latitude'])
                    long=ut.xcr_to_long(child.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}longitude'])
                    alt=ut.xcr_to_alt(child.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}altitude'])
                    self.geospatialTransform=np.array([lat, long, alt])
                if 'Coordinates' in attribute:
                    self.coordinateSystem=child.attrib[attribute]
                if 'FocalLength35mm' in attribute:
                    self.focalLength35mm=ut.xml_to_float(child.attrib[attribute])
                if 'PrincipalPointU' in attribute:
                    self.principalPointU=ut.xml_to_float(child.attrib[attribute])
                if 'PrincipalPointV' in attribute:
                    self.principalPointV=ut.xml_to_float(child.attrib[attribute])

            #Nodes
            rotationnode=child.find('{http://www.capturingreality.com/ns/xcr/1.1#}Rotation')
            rotation=None
            if rotationnode is not None:
                rotation=ut.string_to_rotation_matrix(rotationnode.text)

            positionnode=child.find('{http://www.capturingreality.com/ns/xcr/1.1#}Position')
            translation=None
            if positionnode is not None:
                translation=np.asarray(ut.string_to_list(positionnode.text))
             
            self.cartesianTransform=gt.get_cartesian_transform(translation=translation,rotation=rotation)
            
            coeficientnode=child.find('{http://www.capturingreality.com/ns/xcr/1.1#}DistortionCoeficients')
            if coeficientnode is not None:
                self.distortionCoeficients=ut.string_to_list(coeficientnode.text)  
        return True   

    def get_metadata_from_xml_path(self) ->bool:
        """Extract image metadata from XML Node generated by Agisoft Metashape (self.xmlData and self.subject should be present).

        Features:
            1. cartesianTransform (np.array(4x4))\n
            2. sxy: accuracy in XY (m)\n
            3. sz: accuracy in Z (m) \n

        Returns:
            bool: True if metadata is sucesfully parsed
        """
        if self.xmlPath is None or not os.path.exists(self.xmlPath):
            return False

        if (getattr(self,'cartesianTransform',None) is not None and
            getattr(self,'sxy',None) is not None and
            getattr(self,'sz',None) is not None ):
            return True
        
        self.timestamp=ut.get_timestamp(self.xmlPath)        
        mytree = ET.parse(self.xmlPath)
        root = mytree.getroot()          
        xmlNode = next(cam for cam in root.findall('.//camera') if (cam.get('label') ==self.name or cam.get('label') ==ut.get_subject_name(self.subject) ))
        
        if xmlNode:
            #AGISOFT PARSING 1
            for child in xmlNode.iter('reference'):  
                #get translation
                x =  child.get('x')
                y =  child.get('y')
                z =  child.get('z')
                if x and y and z:
                    translation=np.array([float(x),float(y),float(z)])
                    self.cartesianTransform= gt.get_cartesian_transform(translation=translation)
                #get rotations
                yaw =  child.get('yaw')
                pitch =  child.get('pitch')
                roll =  child.get('roll')
                if yaw and pitch and roll:
                    rotation = gt.get_rotation_matrix(np.array([float(yaw),float(pitch),float(roll)]))
                    self.cartesianTransform=gt.get_cartesian_transform(translation=translation, rotation=rotation)
                #get accuracies
                sxy =  child.get('sxy')
                if sxy:
                    self.sxy=float(sxy)
                sz =  child.get('sz')
                if sz:
                    self.sz=float(sz)
            
            #AGISOFT PARSING 2
            transform=xmlNode.find('transform')
            if transform is not None:
                self.cartesianTransform=ut.string_to_list(transform.text)
        else:
            raise ValueError ('subject not in xml file') 

    def set_cartesianTransform(self,value):
        """Set the cartesianTransform of the ImageNode from various inputs.
        
        Args:
            1. cartesianTransform(np.ndarray(4x4))\n
            2. np.ndarray or Vector3dVector (1x3)  \n
            3. cartesianBounds (np.ndarray (6x1))\n
            4. np.ndarray or Vector3dVector (8x3 or nx3)\n
            5. Open3D.geometry
        """        
        try: #np.ndarray (4x4) 
            self._cartesianTransform=np.reshape(value,(4,4))
        except:
            try: #np.ndarray or Vector3dVector (1x3)  
                self._cartesianTransform=gt.get_cartesian_transform(translation=np.asarray(value))
            except:  
                try: # cartesianBounds (np.ndarray (6x1))
                    self._cartesianTransform=gt.get_cartesian_transform(cartesianBounds=np.asarray(value))
                except:
                    try: # np.ndarray or Vector3dVector (8x3 or nx3)
                        center=np.mean(np.asarray(value),0)
                        self._cartesianTransform=gt.get_cartesian_transform(translation=center)
                    except:
                        try: # Open3D.geometry
                            self._cartesianTransform=gt.get_cartesian_transform(translation=value.get_center())
                        except:
                            raise ValueError('Input must be np.ndarray(6x1,4x4,3x1,nx3), an Open3D geometry or a list of Vector3dVector objects.')


    def get_cartesian_transform(self) -> np.ndarray:
        """Get the cartesianTransform from various inputs.
        
        Args:
            1. self.cartesianBounds (np.array(6x1))  \n
            2. self.orientedBounds (np.array(8x3)) or a list of Vector3dVector objects  \n
            3. orientedBoundingBox\n
            4. Open3D.geometry

        Returns:
            cartesianTransform(np.ndarray(4x4))
        """
        if self._cartesianTransform is not None:
            pass
        elif getattr(self,'_cartesianBounds',None) is not None:
            self._cartesianTransform=gt.get_cartesian_transform(cartesianBounds=self._cartesianBounds)
        elif getattr(self,'_orientedBounds',None) is not None:
            center=np.mean(self._orientedBounds,0)
            self._cartesianTransform=gt.get_cartesian_transform(translation=center)
        elif getattr(self,'_orientedBoundingBox',None) is not None:
            self._cartesianTransform=gt.get_cartesian_transform(translation=self._orientedBoundingBox.get_center())
        elif self._resource is not None:
            self._cartesianTransform=gt.get_cartesian_transform(translation=self._resource.get_center())
        else:
            return None
        return self._cartesianTransform


import copy
import math
import os
import shutil
import time
import unittest
from datetime import datetime
from multiprocessing.sharedctypes import Value

import cv2
import geomapi.utils as ut
import ifcopenshell
import numpy as np
import open3d as o3d
import PIL
import pye57
from geomapi.nodes.bimnode import BIMNode
from geomapi.nodes.imagenode import ImageNode
from geomapi.nodes.meshnode import MeshNode
from geomapi.nodes.orthonode import OrthoNode
from geomapi.nodes.pointcloudnode import PointCloudNode
from ifcopenshell.util.selector import Selector
from rdflib import RDF, RDFS, XSD, Graph, Literal, URIRef


class TestUtils(unittest.TestCase):
 ################################## SETUP/TEARDOWN CLASS ######################
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") #os.pardir, 
        
        #GRAPH 1
        cls.graphPath1=os.path.join(cls.path,'bimGraph1.ttl')
        cls.graph1=Graph().parse(cls.graphPath1)

        #GRAPH 2
        cls.graphPath2=os.path.join(cls.path,'resourceGraph.ttl')
        cls.graph2=Graph().parse(cls.graphPath2)

        #GRAPH 3
        cls.graphPath3=os.path.join(cls.path,'pcdGraph.ttl')
        cls.graph3=Graph().parse(cls.graphPath3)

        #GRAPH 4
        cls.graphPath4=os.path.join(cls.path,'meshGraph.ttl')
        cls.graph4=Graph().parse(cls.graphPath4)
        
        #GRAPH 5
        cls.graphPath5=os.path.join(cls.path,'imgGraph.ttl')
        cls.graph5=Graph().parse(cls.graphPath5)
        
        #IMG
        cls.image1Path=os.path.join(cls.path,'IMG',"IMG_2173.JPG")  
        cls.image1=cv2.imread(cls.image1Path)
        cls.image1CartesianTransform= np.array([[-0.0544245051021791, 0.089782180920334, 0.994473294379276, -8.94782374621677],
                                                [-0.783686718502243, -0.621016494566922, 0.0131772804097903 ,11.2531401937057],
                                                [0.618767404189627, -0.778638345745315, 0.104159618122526, 6.5428452363933],
                                                [0,0,0,1]])
        cls.image2Path=os.path.join(cls.path,'IMG',"IMG_2174.JPG")  
        cls.image2=cv2.imread(cls.image1Path)
        cls.image2CartesianTransform= np.array([[-0.046509031201878, 0.0485391010476459, 0.99773787423659, -8.63657982153356],
                                                [-0.714937318920729, -0.699188212147553, 0.00068848264200394 ,9.21354145067747],
                                                [0.697639978807912, -0.713288020131696, 0.0672209811405822,6.57082854991429],
                                                [0,0,0,1]])
        cls.imageDronePath=os.path.join(cls.path,'IMG',"DJI_0067.JPG")  
        
        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)
   
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')


    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        shutil.rmtree(cls.resourcePath)      
################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))


################################## TEST FUNCTIONS ######################
    def test_bind_ontologies(self):
        graph=Graph()
        graph=ut.bind_ontologies(graph) 
        test=False
        for n in graph.namespaces():
            if 'v4d' in n:
                test=True
        self.assertTrue(test)

    def test_cartesianTransform_to_literal(self):
        literal=ut.cartesianTransform_to_literal(self.image1CartesianTransform)
        self.assertTrue(str(-0.0544) in literal)
       
    def test_check_if_uri_exists(self):
        list=[URIRef('4499de21-f13f-11ec-a70d-c8f75043ce59'),URIRef('http://IMG_2173'),URIRef('http://000_GM_Opening_Rectangular_Opening_Rectangular_1101520'),URIRef('43be9b1c-f13f-11ec-8e65-c8f75043ce59')]
        subject=URIRef('43be9b1c-f13f-11ec-8e65-c8f75043ce59')
        test=ut.check_if_uri_exists(list, subject)
        self.assertTrue(test)

        #incorrect one
        subject=URIRef('blablabla')
        test=ut.check_if_uri_exists(list, subject)
        self.assertFalse(test)
       
    def test_validate_timestamp(self):

        #string
        self.assertEqual(str(ut.validate_timestamp("2022:03:13 13:55:26")),"2022-03-13T13:55:26")

        #string
        self.assertEqual(str(ut.validate_timestamp('Tue Dec  7 09:38:13 2021')),"2021-12-07T09:38:13")

         #string
        self.assertEqual(str(ut.validate_timestamp("1648468136.033126")),"2022-03-28T13:48:56")

        #tuple
        self.assertEqual(str(ut.validate_timestamp(datetime(2022,3,13,13,55,26))),"2022-03-13T13:55:26")

        # invalid
        self.assertRaises(ValueError,ut.validate_timestamp,'qsdfqsdf')

    def test_validate_string(self):
        
        test=ut.validate_string('http://pointcloud2_0')
        self.assertEqual(test,'http://pointcloud2_0')
        test=ut.validate_string('http://po/int/cloud2_ 0:')
        self.assertEqual(test,'http://po_int_cloud2__0_')
        test=ut.validate_string('pointc&:lo ud2_0_')
        self.assertEqual(test,'pointc&_lo_ud2_0_')
        test=ut.validate_string('file:///pointc&:lo ud2_0')
        self.assertEqual(test,'file:///pointc&_lo_ud2_0')
        test=ut.validate_string('file:///pointc&:lo /ud2_0')
        self.assertEqual(test,'file:///pointc&_lo__ud2_0')
        test=ut.validate_string('4499de21-f13f-11ec-a70d-c8f75043ce59')
        self.assertEqual(test,'4499de21-f13f-11ec-a70d-c8f75043ce59')
        test=ut.validate_string('[this<has$to^change]')
        self.assertEqual(test,'_this_has_to_change_')

    def test_get_node_resource_extensions(self):
        self.assertEqual(len(ut.MESH_EXTENSION),len(ut.get_node_resource_extensions(str(type(MeshNode())))))
        self.assertEqual(len(ut.MESH_EXTENSION),len(ut.get_node_resource_extensions(str(type(BIMNode())))))
        self.assertEqual(len(ut.PCD_EXTENSION),len(ut.get_node_resource_extensions(str(type(PointCloudNode())))))
        self.assertEqual(len(ut.IMG_EXTENSION),len(ut.get_node_resource_extensions(str(type(ImageNode())))))
        self.assertEqual(len(ut.IMG_EXTENSION),len(ut.get_node_resource_extensions(str(type(OrthoNode())))))
        self.assertEqual(len(ut.MESH_EXTENSION),len(ut.get_node_resource_extensions(str(type(MeshNode())))))

    def test_clean_attributes_list(self):
        list=['blablabla',
            'graph','graphPath','subject','fullResourcePath','kwargs', 'orientedBoundingBox',
            'ifcElement',
            'mesh',
            'exifData','xmlData','image','features2d','pinholeCamera',
            'pcd','e57Pointcloud','e57xmlNode','e57image','features3d',
            'linkedNodes']
        newList=ut.clean_attributes_list(list)
        self.assertEqual(len(newList),1)
      
    def test_dd2dms(self):
        dms=ut.dd2dms(6.5)
        self.assertEqual(dms[1],30)

    def test_dms2dd(self):
        #west
        dd=ut.dms2dd(6, 30, 0, 'W')
        self.assertEqual(dd,-6.5)
        #south
        dd=ut.dms2dd(6, 30, 0, 'S')
        self.assertEqual(dd,-6.5)
        #north
        dd=ut.dms2dd(6, 30, 0, 'N')
        self.assertEqual(dd,6.5)
      
    def test_filter_exif_gps_data(self):
        #tuple
        dd=ut.filter_exif_gps_data((6, 30, 0), reference= 'N') 
        self.assertEqual(dd,6.5)

        #float
        dd=ut.filter_exif_gps_data(6.5, reference= 'S') 
        self.assertEqual(dd,-6.5)
        
    def test_get_attribute_from_predicate(self):
        graph=Graph()
        graph=ut.bind_ontologies(graph)
        string=ut.get_attribute_from_predicate(graph, predicate =Literal('http://libe57.org#pointCount')) 
        self.assertEqual(string,'pointCount')
        string=ut.get_attribute_from_predicate(graph, predicate =Literal('https://w3id.org/v4d/core#faceCount')) 
        self.assertEqual(string,'faceCount')
        string=ut.get_attribute_from_predicate(graph, predicate =Literal('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#timestamp')) 
        self.assertEqual(string,'timestamp')
        string=ut.get_attribute_from_predicate(graph, predicate =Literal('http://ifcowl.openbimstandards.org/IFC2X3_Final#className')) 
        self.assertEqual(string,'className')
        string=ut.get_attribute_from_predicate(graph, predicate =Literal('http://www.w3.org/2003/12/exif/ns#xResolution')) 
        self.assertEqual(string,'xResolution')
        string=ut.get_attribute_from_predicate(graph, predicate =Literal('https://w3id.org/gom#coordinateSystem')) 
        self.assertEqual(string,'coordinateSystem')
      
    def test_get_exif_data(self):
        im = PIL.Image.open(self.imageDronePath) 
        exifData=ut.get_exif_data(im)
        self.assertIsNotNone(exifData["GPSInfo"])
        im.close()
       
    def test_get_extension(self):
        string=ut.get_extension(self.image1Path)
        self.assertEqual(string,'.JPG')

    def test_get_filename(self):
        string=ut.get_filename(self.image1Path)
        self.assertEqual(string,'IMG_2173')

    def test_get_folder(self):
        string=ut.get_folder(self.image1Path)
        self.assertEqual(string,os.path.join(self.path,'IMG'))

    def test_get_folder_path(self):
        string=ut.get_folder_path(self.image1Path)
        self.assertEqual(string,os.path.join(self.path,'IMG'))

    def test_get_if_exist(self):
        im = PIL.Image.open(self.imageDronePath) 
        exifData=ut.get_exif_data(im)
        self.assertIsNotNone(ut.get_if_exist(exifData,"XResolution"))
        self.assertIsNone(ut.get_if_exist(exifData,"qsdfsdf"))

    def test_get_list_of_files(self):
        files=ut.get_list_of_files(self.path)
        self.assertEqual(len(files),385)

    def test_get_paths_in_class(self):
        class tinyClass:
            def __init__(self,**kwargs):
                self.__dict__.update(kwargs)  

        test=tinyClass(path=1,myPath=2,thisisnotapat=3,resolution=4)
        paths=ut.get_paths_in_class(test)
        self.assertEqual(len(paths),2)

    def test_get_subject_graph(self):
        
        # graph=Graph().parse(self.graphPath,format='ttl')
        subject=URIRef('http://000_GM_Opening_Rectangular_Opening_Rectangular_1101520')

        #subject
        subject=next(s for s in self.graph1.subjects())
        newGraph=ut.get_subject_graph(graph=self.graph1,subject=subject)
        self.assertEqual(len(newGraph),len([t for t in self.graph1.triples((subject,None,None))]))

        #no subject        
        newGraph=ut.get_subject_graph(graph=self.graph2)
        self.assertIsNotNone(next(newGraph.subjects(RDF.type)))

        #wrong subject
        self.assertRaises(ValueError,ut.get_subject_graph,graph=self.graph5,subject=URIRef('blabla'))

    def test_get_timestamp(self):
        timeStamp=ut.get_timestamp(self.image1Path) 
        self.assertEqual(timeStamp,'2022-08-02T08:25:00')

    def test_get_variables_in_class(self):
        class tinyClass:
            def __init__(self,**kwargs):
                self.__dict__.update(kwargs)  

        test=tinyClass(path=1,myPath=2,thisisnotapat=3,resolution=4)
        variables=ut.get_variables_in_class(test)
        self.assertEqual(len(variables),4)

    def test_is_float(self):
        test=ut.is_float(0.5) 
        self.assertTrue(test)

        test=ut.is_float('0.5') 
        self.assertTrue(test)

        test=ut.is_float('qsfdsdf') 
        self.assertFalse(test)

    def test_is_int(self):
        test=ut.is_int(1) 
        self.assertTrue(test)

        test=ut.is_int('1') 
        self.assertTrue(test)

        test=ut.is_int('qsfdsdf') 
        self.assertFalse(test)

        test=ut.is_int(0.5) 
        self.assertTrue(test)
        
    def test_is_string(self):
        test=ut.is_string(1) 
        self.assertTrue(test)

        test=ut.is_string('1') 
        self.assertTrue(test)

        test=ut.is_string('qsfdsdf') 
        self.assertTrue(test)

        test=ut.is_string(0.5) 
        self.assertTrue(test)
            
    def test_is_uriref(self):
        test=ut.is_uriref(Literal(1)) 
        self.assertTrue(test)

        test=ut.is_uriref('1') 
        self.assertTrue(test)

        test=ut.is_uriref('qsfdsdf') 
        self.assertTrue(test)

        test=ut.is_uriref(0.5) 
        self.assertFalse(test)

    def test_item_to_list(self):
        item='qsdf'
        test=ut.item_to_list(item)
        self.assertEqual(len(test),1)
        item=['qsdf']
        test=ut.item_to_list(item)
        self.assertEqual(len(test),1)
        item=['qsdf',1]
        test=ut.item_to_list(item)
        self.assertEqual(len(test),2)

    def test_literal_to_array(self):
        #cartesianBounds
        item=Literal("[-12.33742784 -10.91544131  73.8353109   73.96926636   8.642    9.462     ]")
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,6)

        item=Literal([-12.33742784, -10.91544131,  73.8353109 ,  73.96926636 ,  8.642 ,   9.462     ])
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,6)

        item=Literal(np.array([-12.33742784, -10.91544131,  73.8353109 ,  73.96926636 ,  8.642 ,   9.462     ]))
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,6)

        item=Literal("[-12.33742784, -10.91544131,  73.8353109  , 73.96926636  , 8.642   , 9.462     ]")
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,6)

        item=Literal("None")
        test=ut.literal_to_array(item)
        self.assertEqual(test,None)

        #geospatialTransform
        item=Literal("[6.30  5  0]" )
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,3)
        
        item=Literal("[6.30 , 5 , 0]" )
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,3)

        item=Literal([6.30 , 5 , 0])
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,3)

        item=Literal(np.array([6.30 , 5 , 0]) )
        test=ut.literal_to_array(item)
        self.assertEqual(test.size,3)
        
        item=Literal("[None, None, None]" )
        test=ut.literal_to_array(item)
        self.assertEqual(test,None)

    def test_literal_to_linked_subjects(self):
        string="['file:///Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1095339', 'file:///IMG_2174', 'file:///week22_photogrammetry_-_Cloud', 'file:///Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1118860_0KysUSO6T3_gOJKtAiUE7d']"
        list=ut.literal_to_linked_subjects(string)
        gtlist= ['file:///Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1095339', 'file:///IMG_2174', 'file:///week22_photogrammetry_-_Cloud', 'file:///Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1118860_0KysUSO6T3_gOJKtAiUE7d']
        (self.assertTrue(list[i]==gtlist[i]) for i in range(len(list)) )

    def test_check_if_subject_is_in_graph(self):
        #http
        self.assertTrue(ut.check_if_subject_is_in_graph(self.graph2,next(s for s in self.graph2.subjects(RDF.type))))

        #file
        self.assertTrue(ut.check_if_subject_is_in_graph(self.graph5,next(s for s in self.graph5.subjects(RDF.type))))

        #random
        graph=Graph()
        graph.add((URIRef('mySubject'),RDFS.label,Literal('label')))
        self.assertTrue(ut.check_if_subject_is_in_graph(graph,URIRef('mySubject')))

        #not in graph
        self.assertFalse(ut.check_if_subject_is_in_graph(self.graph4,URIRef('ikjuhygfds')))

    def test_get_graph_subject(self):
        #http
        self.assertIsNotNone(ut.get_graph_subject(self.graph1,next(s for s in self.graph1.subjects(RDF.type))))
        
        #file
        self.assertIsNotNone(ut.get_graph_subject(self.graph3,next(s for s in self.graph3.subjects(RDF.type))))

        #random
        graph=Graph()
        graph.add((URIRef('mySubject'),RDFS.label,Literal('label')))
        self.assertIsNotNone(ut.get_graph_subject(graph,URIRef('mySubject')))

        #not in graph
        self.assertRaises(ValueError,ut.get_graph_subject,self.graph4,URIRef('kjhgfd'))

    def test_get_data_type(self):
        value=1
        dataType=ut.get_data_type(value)
        self.assertAlmostEqual(dataType,XSD.integer)
        
        value=0.1
        dataType=ut.get_data_type(value)
        self.assertAlmostEqual(dataType,XSD.float)

        value=[1,2,3]
        dataType=ut.get_data_type(value)
        self.assertAlmostEqual(dataType,XSD.string)

        value=datetime(1991,5,12,10,10,10)
        dataType=ut.get_data_type(value)
        self.assertAlmostEqual(dataType,XSD.dateTime)
        
        value=np.array([1,2,3])
        dataType=ut.get_data_type(value)
        self.assertAlmostEqual(dataType,XSD.string)

        value=(1,2,3)
        dataType=ut.get_data_type(value)
        self.assertAlmostEqual(dataType,XSD.string)

    def test_literal_to_cartesianTransform(self):
        item=Literal("[[-0.05442451  0.08978218  0.99447329 -8.94782375] [-0.78368672 -0.62101649  0.01317728 11.25314019] [ 0.6187674  -0.77863835  0.10415962  6.54284524] [ 0.          0.          0.          1.        ]]")
        test=ut.literal_to_cartesianTransform(item)
        self.assertEqual(test.size,16)

        item=Literal('[[-0.05442451  0.08978218  0.99447329 -8.94782375]\r\n [-0.78368672 -0.62101649  0.01317728 11.25314019]\r\n [ 0.6187674  -0.77863835  0.10415962  6.54284524]\r\n [ 0.          0.          0.          1.        ]]')
        test=ut.literal_to_cartesianTransform(item)
        self.assertEqual(test.size,16)
        
        item=Literal([-0.05442451,  0.08978218 , 0.99447329 ,-8.94782375,-0.78368672, -0.62101649 , 0.01317728, 11.25314019, 0.6187674 , -0.77863835 , 0.10415962 , 6.54284524, 0.  ,        0.    ,      0.     ,     1.        ])
        test=ut.literal_to_cartesianTransform(item)
        self.assertEqual(test.size,16)

        item=Literal(np.array([[-0.05442451,  0.08978218 , 0.99447329 ,-8.94782375],
                                 [-0.78368672, -0.62101649 , 0.01317728, 11.25314019],
                                  [ 0.6187674 , -0.77863835 , 0.10415962 , 6.54284524] ,
                                  [ 0.  ,        0.    ,      0.     ,     1.        ]]))
        test=ut.literal_to_cartesianTransform(item)
        self.assertEqual(test.size,16)

        item=Literal("[[-0.05442451 , 0.08978218 , 0.99447329 ,-8.94782375] [-0.78368672, -0.62101649,  0.01317728, 11.25314019] [ 0.6187674  ,-0.77863835 , 0.10415962 , 6.54284524] [ 0.  ,        0.    ,      0.       ,   1.        ]]")
        test=ut.literal_to_cartesianTransform(item)
        self.assertEqual(test.size,16)

        item=Literal("None")
        test=ut.literal_to_cartesianTransform(item)
        self.assertIsNone(test)

    def test_literal_to_float(self):
        item=Literal(0.5)
        test=ut.literal_to_float(item)
        self.assertIsInstance(test,float)

        item=Literal(5)
        test=ut.literal_to_float(item)
        self.assertIsInstance(test,float)

        item=Literal('0.5')
        test=ut.literal_to_float(item)
        self.assertIsInstance(test,float)

        item=Literal('blabla')
        self.assertRaises(ValueError,ut.literal_to_float,item)

    def test_literal_to_int(self):
        item=Literal(0.5)
        test=ut.literal_to_int(item)
        self.assertIsInstance(test,int)

        item=Literal(5)
        test=ut.literal_to_int(item)
        self.assertIsInstance(test,int)

        item=Literal('5')
        test=ut.literal_to_int(item)
        self.assertIsInstance(test,int)

        item=Literal('blabla')
        self.assertRaises(ValueError,ut.literal_to_int,item)

    def test_literal_to_list(self):
        item=Literal(0.5)
        test=ut.literal_to_list(item)
        self.assertIsInstance(test,list)

        item=Literal('[-0.126115439984335, 0.0981832072267781, 0.0312044509604729]')
        test=ut.literal_to_list(item)
        self.assertEqual(len(test),3)

        item=Literal('[-0.126115439984335 0.0981832072267781 0.0312044509604729]')
        test=ut.literal_to_list(item)
        self.assertEqual(len(test),3)

        item=Literal('[None None ]')
        test=ut.literal_to_list(item)
        self.assertIsNone(test)

    def test_literal_to_orientedBounds(self):
        item=Literal("[[-0.05442451  0.08978218  0.99447329 ] [-0.78368672 -0.62101649  0.01317728 ] [ 0.6187674  -0.77863835    6.54284524] [ 0.          0.                    1.        ] [-0.05442451   0.99447329 -8.94782375] [-0.78368672  0.01317728 11.25314019] [   -0.77863835  0.10415962  6.54284524] [ 0.                    0.          1.        ]]")
        test=ut.literal_to_orientedBounds(item)
        self.assertEqual(test.size,24)

        item=Literal([-0.05442451,  0.08978218 , 0.99447329 ,-8.94782375,-0.78368672, -0.62101649 , 0.01317728, 11.25314019, 0.6187674 , -0.77863835 , 0.10415962 , 6.54284524, 0.  ,        0.    ,      0.  , -0.77863835 , 0.10415962 , 6.54284524, -0.77863835 , 0.10415962 , 6.54284524, -0.77863835 , 0.10415962 , 6.54284524        ])
        test=ut.literal_to_orientedBounds(item)
        self.assertEqual(test.size,24)

        item=Literal(np.array([[-0.05442451,  0.08978218 , 0.99447329 ],
                                 [-0.78368672, -0.62101649 , 0.01317728],
                                  [ 0.6187674 , -0.77863835 , 0.10415962 ] ,
                                  [ 0.  ,        0.    ,      0.     ,          ],
                                  [-0.05442451,  0.08978218 , 0.99447329 ],
                                 [-0.78368672, -0.62101649 , 0.01317728],
                                  [ 0.6187674 , -0.77863835 , 0.10415962 ] ,
                                  [ 0.  ,        0.    ,      0.     ,          ]]))
        test=ut.literal_to_orientedBounds(item)
        self.assertEqual(test.size,24)

        item=Literal("[[-0.05442451 , 0.08978218 , 0.99447329 ] [-0.78368672, -0.62101649 , 0.01317728 ] [ 0.6187674  ,-0.77863835  ,  6.54284524] [ 0.    ,      0.          ,          1.        ] [-0.05442451  , 0.99447329 ,-8.94782375] [-0.78368672 , 0.01317728, 11.25314019] [   -0.77863835,  0.10415962 , 6.54284524] [ 0.       ,             0.   ,       1.        ]]")
        test=ut.literal_to_orientedBounds(item)
        self.assertEqual(test.size,24)

        item=Literal("None")
        test=ut.literal_to_orientedBounds(item)
        self.assertIsNone(test)

    def test_literal_to_string(self):
        item=Literal(0.5)
        test=ut.literal_to_string(item)
        self.assertIsInstance(test,str)

        item=Literal(5)
        test=ut.literal_to_string(item)
        self.assertIsInstance(test,str)

        item=Literal('blabla')
        test=ut.literal_to_string(item)
        self.assertIsInstance(test,str)

        item=Literal('None')
        test=ut.literal_to_string(item)
        self.assertIsNone(test)

    def test_literal_to_uriref(self):
        item=Literal('blabla')
        test=ut.literal_to_uriref(item).toPython()
        self.assertEqual(test,'blabla')

        item=Literal('None')
        test=ut.literal_to_uriref(item)
        self.assertIsNone(test)

        self.assertRaises(ValueError,ut.literal_to_uriref,5)

        self.assertRaises(ValueError,ut.literal_to_uriref,0.5)

    def test_match_uri(self):
        test=ut.match_uri('timestamp').toPython()
        self.assertEqual(test,'https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#timestamp')
        test=ut.match_uri('cartesianBounds').toPython()
        self.assertEqual(test,'http://libe57.org#cartesianBounds')
        test=ut.match_uri('coordinateSystem').toPython()
        self.assertEqual(test,'https://w3id.org/gom#coordinateSystem')
        test=ut.match_uri('ifcPath').toPython()
        self.assertEqual(test,'http://ifcowl.openbimstandards.org/IFC2X3_Final#ifcPath')
        test=ut.match_uri('xResolution').toPython()
        self.assertEqual(test,'http://www.w3.org/2003/12/exif/ns#xResolution')
        test=ut.match_uri('focalLength35mm').toPython()
        self.assertEqual(test,'http://www.w3.org/1999/02/22-rdf-syntax-ns#focalLength35mm')
        test=ut.match_uri('qfdsfqsf').toPython()
        self.assertEqual(test,'https://w3id.org/v4d/core#qfdsfqsf')

    def test_parse_dms(self):
        item="[[6 , 30 , 0, N ]]"
        test=ut.parse_dms(item)
        self.assertEqual(test,6.5)

        item="[6  30  0 N ]"
        test=ut.parse_dms(item)
        self.assertEqual(test,6.5)

        item=(6 , 30 , 0 , 'N' )
        test=ut.parse_dms(item)
        self.assertEqual(test,6.5)

        item=np.array([6 , 30 , 0 , 'N'])
        test=ut.parse_dms(item)
        self.assertEqual(test,6.5)
    
    def test_replace_str_index(self):
        item="rrrr"
        test=ut.replace_str_index(item,index=0,replacement='_')
        self.assertEqual(test[0],'_')

    def test_string_to_list(self):
        #cartesianBounds
        item="[-12.33742784 -10.91544131  73.8353109   73.96926636   8.642    9.462     ]"
        test=ut.string_to_list(item)
        self.assertEqual(len(test),6)

        item="[-12.33742784, -10.91544131,  73.8353109  , 73.96926636  , 8.642   , 9.462     ]"
        test=ut.string_to_list(item)
        self.assertEqual(len(test),6)

        item="None"
        test=ut.string_to_list(item)
        self.assertIsNone(test)
        
        item="[None, None, None]" 
        test=ut.string_to_list(item)
        self.assertIsNone(test)

    def test_string_to_rotation_matrix(self):
        #cartesianBounds
        item="[-12.33742784 -10.91544131  73.8353109   73.96926636   8.642    9.462 -12.33742784 -10.91544131  73.8353109    ]"
        test=ut.string_to_rotation_matrix(item)
        self.assertEqual(test.size,9)

        item="[-12.33742784, -10.91544131,  73.8353109  , 73.96926636  , 8.642   , 9.462  ,-12.33742784, -10.91544131,  73.8353109   ]"
        test=ut.string_to_rotation_matrix(item)
        self.assertEqual(test.size,9)

        item="[[-12.33742784, -10.91544131,  73.8353109 ] ,   [ 73.96926636  , 8.642   , 9.462  ] ,  [-12.33742784, -10.91544131,  73.8353109   ]]"
        test=ut.string_to_rotation_matrix(item)
        self.assertEqual(test.size,9)

        item="None"
        self.assertRaises(ValueError,ut.string_to_rotation_matrix,item)
        
        item="[[None, None, None],[None, None, None],[None, None, None]]"     
        self.assertRaises(ValueError,ut.string_to_rotation_matrix,item)

    def test_xcr_to_alt(self):
        item="642069440/10000"
        test=ut.xcr_to_alt(item)
        self.assertEqual(test,64206.9440)

        item="None"
        test=ut.xcr_to_alt(item)
        self.assertIsNone(test)
        
    def test_xcr_to_lat(self):
        item="179.992700159232641N"
        test=ut.xcr_to_lat(item)
        self.assertEqual(test,179.992700159232641)

        item="179.992700159232641S"
        test=ut.xcr_to_lat(item)
        self.assertEqual(test,-179.992700159232641)

        item="None"
        test=ut.xcr_to_lat(item)
        self.assertIsNone(test)

    def test_xcr_to_long(self):
        item="66.587349536158328E"
        test=ut.xcr_to_long(item)
        self.assertEqual(test,66.587349536158328)

        item="66.587349536158328W"
        test=ut.xcr_to_long(item)
        self.assertEqual(test,-66.587349536158328)

        item="None"
        test=ut.xcr_to_long(item)
        self.assertIsNone(test)

if __name__ == '__main__':
    unittest.main()



"""Tools to complete and combine meshes and pointclouds
"""
import geomapi
import geomapi.tools.completiontools.utils as ut

import open3d as o3d
import numpy as np

from typing import Tuple, List

# The main function to combine 2 aligned geometries
def combine_geometry(ogGeometry: o3d.geometry.Geometry, newGeometry :o3d.geometry.Geometry, distanceTreshold : float = 0.05) -> o3d.geometry.Geometry:
    """Combines 2 aligned geometries assuming the ogGeometry is the reference and the newGeometry will suplement it.
    this is performed in a few steps:
        1) Create a convex hull of the newGeometry
        2) Filter the relevant points of the ogGeometry
        3) Perform a 2 step relevance check on the ogPoints if they fail either, they will be removed
            - Coverage check: perform a distance query, points that to far away from the mesh could be either out of date or not scanned
            - Visibility check: points that are inside the mesh are considered invisible and are kept, visible points are deemed out of date and are removed.
        4) Filter the newPoints to only add the points that are changed
        5) combine the changed-newPoints, the invisible-not-covered-ogPoints, the covered-ogPoints and the irrelevant points

    Args:
        ogGeometry (o3d.geometry): The reference geometry to be completed
        newGeometry (o3d.geometry): The new geometry to be addad
        distanceTreshold (float, optional): The treshold the distance-filtering is set to. Defaults to 0.05.

    Returns:
        o3d.geometry: the combined geometry
    """

    # Step 1: Create a convex hull of the newGeometry
    newGeoHull = get_convex_hull(newGeometry)
    print("Covex hull created")
    # Step 2: Filter out the irrelevant points in the ogGeometry
    relevantOg, irrelevantOg = get_points_in_hull(ogGeometry, newGeoHull)
    print("Irrelevant points filtered")
    # Step 3: Isolate the not covered points of the ogGeometry compared to the newGeometry
    newGeometryPoints = newGeometry.sample_points_poisson_disk(number_of_points=100000)
    coveredPoints, unCoveredPoints = ut.filter_pcd_by_distance(relevantOg, newGeometryPoints, distanceTreshold)
    print("Covered poinys calculated")
    # Step 4: Perform the visibility check of the not covered points
    invisibleUncoveredPoints = get_invisible_points(unCoveredPoints, newGeometry)
    print("invisible points detected")
    # Step 5: Filter the newGeometryPoints to only keep the changed geometry
    existingNewGeo, newNewGeo = ut.filter_pcd_by_distance(newGeometryPoints, relevantOg, distanceTreshold)
    print("new points filtered")
    # Step 6: Combine the irrelevant, unchanged and changed geometry
    newCombinedGeometry = irrelevantOg + coveredPoints + invisibleUncoveredPoints + newNewGeo
    print("geometries combined")
    return newCombinedGeometry

# Converts a geometry to a covex hull
def get_convex_hull(geometry : o3d.geometry.Geometry) ->  o3d.geometry.Geometry:
    """Calculates a convex hull of a generic geometry

    Args:
        geometry (open3d.geometry): The geometry to be hulled

    Returns:
        open3d.geometry: The resulting hull
    """

    hull, _ = geometry.compute_convex_hull()
    return hull

# Returns a filtered mpcd with only points which are inside the convex hull
def get_points_in_hull(geometry : o3d.geometry.Geometry, hull: o3d.geometry.Geometry) -> Tuple[o3d.geometry.Geometry,o3d.geometry.Geometry]:
    """Separates a geometry in points inside and outside a convex hull

    Args:
        geometry (open3d.geometry): The geometry to be filtered
        hull (open3d.geometry): The hull to filter

    Returns:
       Tuple[open3d.geometry, open3d.geometry]: Points inside the hull, Points outside the hull
    """

    hullVerts = np.asarray(hull.vertices)
    points = np.asarray(geometry.points)
    idxs = ut.get_indices_in_hull(points, hullVerts)
    pcdInHull = geometry.select_by_index(idxs)
    pcdOutHull = geometry.select_by_index(idxs, True)
    return pcdInHull, pcdOutHull

# checks if the points is inside the (closed) mesh
def check_point_inside_mesh(points: List, mesh :o3d.geometry.Geometry) -> Tuple[List, List]:
    """Performs a visibility check to chack if the points are inside the mesh or not

    Args:
        points (List): The points to check
        mesh (open3d.geometry): The mesh to check against

    Returns:
        Tuple[List, List]: The points withing the mesh, the points outside the mesh
    """

    # step 0: Set up a raycasting scene
    scene = o3d.t.geometry.RaycastingScene()
    scene.add_triangles(o3d.t.geometry.TriangleMesh.from_legacy(mesh))
    insideList = []
    outsideList = []

    for point in points:
        # step 1: get the closest point to the mesh
        queryPoint = o3d.core.Tensor([point], dtype=o3d.core.Dtype.Float32)
        queryResult = scene.compute_closest_points(queryPoint)
        closestPoint =  queryResult['points'].numpy()
        # step 2 Get the normal of the triangle
        closestTriangle = queryResult['primitive_ids'][0].item()
        triangleNormal = np.asarray(mesh.triangle_normals[closestTriangle])
        # step 3: compare the normal with the ray direction
        closestPointDirection = closestPoint - point
        dotProduct = (closestPointDirection @ triangleNormal)[0]
        if (dotProduct > 0):
            insideList.append(point)
        else:
            outsideList.append(point)

    return insideList, outsideList

# geturns all the points that are inside the (closed) mesh
def get_invisible_points(points : o3d.geometry.Geometry, mesh: o3d.geometry.Geometry) -> o3d.geometry.Geometry:
    """Returns all the points that are not visible because the are within a mesh

    Args:
        points (open3d.geometry): The points to chack
        mesh (open3d.geometry): The mesh to check against

    Returns:
        open3d.geometry: The filtered poincloud
    """
    
    insideList, outsideList = check_point_inside_mesh(points.points, mesh)
    visiblePoints = o3d.geometry.PointCloud()
    visiblePoints.points = o3d.utility.Vector3dVector(outsideList)
    invisiblePoints = o3d.geometry.PointCloud()
    invisiblePoints.points = o3d.utility.Vector3dVector(insideList)
    return invisiblePoints

def sample_mesh_with_density(mesh: o3d.geometry.TriangleMesh, density : float) -> o3d.geometry.PointCloud:
    print("not implemented yet")
    pass


"""
validationtools - a Python library for validating objects.
"""
#IMPORT PACKAGES
from lib2to3.pytree import Node
import numpy as np 
import cv2 
import open3d as o3d 
import json  
import os 
import re
import matplotlib.pyplot as plt #conda install -c conda-forge matplotlib
#import torch #conda install -c pytorch pytorch
# import pye57 #conda install xerces-c  =>  pip install pye57
import xml.etree.ElementTree as ET 
# from pathlib import Path
import math
import xlsxwriter
import csv
import copy
from PIL import Image, ImageDraw
import time 
import ifcopenshell.geom as geom
import math
from scipy.spatial.transform import Rotation as R
from PIL import Image, ImageDraw
import shutil
from rdflib import Graph, URIRef, RDF
import os.path
import importlib
import numpy as np
import xml.etree.ElementTree as ET
import open3d as o3d
import uuid    
import copy
import pye57 
import ifcopenshell
import multiprocessing as mp
import time
from pathlib import Path
from typing import List
import matplotlib.pyplot as plt
import csv
import xlsxwriter

# import ifcopenshell.util
# import ifcopenshell.geom as geom
# from ifcopenshell.util.selector import Selector
# from ifcopenshell.ifcopenshell_wrapper import file

# import APIs
import rdflib
from rdflib import Graph, plugin
from rdflib.serializer import Serializer #pip install rdflib-jsonld https://pypi.org/project/rdflib-jsonld/
from rdflib import Graph
from rdflib import URIRef, BNode, Literal
from rdflib.namespace import CSVW, DC, DCAT, DCTERMS, DOAP, FOAF, ODRL2, ORG, OWL, \
                           PROF, PROV, RDF, RDFS, SDO, SH, SKOS, SOSA, SSN, TIME, \
                           VOID, XMLNS, XSD


import geomapi.utils.geometryutils as gt

#IMPORT MODULES 
from geomapi.nodes import *
from geomapi.nodes.sessionnode import create_node 
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt

def get_boundingbox_of_list_of_geometries(geometries:List[Node]) -> np.array:
    """Determines the global boundingbox of a group of Node containing geometries.

    Args:
        geometries (List[Nodes]):  list of Nodes containing a resource of which the boundingbox must be determined"
            
    Returns:
        np.array[3x1]
    """
    pcd = o3d.geometry.PointCloud()
    for n in geometries:
        n.get_resource()
        if n.resource is not None:
            pcd.__iadd__(o3d.geometry.PointCloud(gt.get_oriented_bounds(gt.get_cartesian_bounds(n.resource))))
    cartesianBounds = gt.get_cartesian_bounds(pcd.get_oriented_bounding_box())
    return cartesianBounds

def optimize_camera_for_overview(geometries:List[Node]):
    """Determines the camera position to create an overview image of a group of Nodes containing geometries.

    Args:
        geometries (List[Nodes]):  list of Nodes containing a resource "
            
    Returns:
        c center of the objects np.array[1x3]
        c_i the camera position np.array[1x3]
        up defines which direction is up [0,0,1] for z
    """

    cartesianBounds = get_boundingbox_of_list_of_geometries(geometries=geometries)
    bbbox = gt.get_oriented_bounding_box(cartesianBounds)

    fov = np.pi / 3 #60 #degrees
            
    #determine extrinsic camera parameters
    extrinsic = np.empty((1,3), dtype=float)
    
    c = bbbox.get_center()
    u = bbbox.extent[0]
    d_w = math.cos(fov/2)*u

    #determine c_i
    rotation_matrix = bbbox.R
    pcd2 = o3d.geometry.PointCloud()
    array = np.array([[c[0],c[1],c[2]+d_w]])
    pcd2.points = o3d.utility.Vector3dVector(array)
    c_i = np.asarray(pcd2.points[0])

    up = [0, 0, 1]  # camera orientation

    
    return c,c_i,up

def create_croppedpcd(pointcloud: PointCloudNode, element: BIMNode, bb: o3d.geometry.OrientedBoundingBox, sn : SessionNode) -> PointCloudNode:
    """Crops a part of a point cloud around a BIM object.

    Args:
        pointcloud (PointCloudNode): Pointcloud from which the zone needs to be cropped
        element (BIMNode): The BIMNode that the cropped region represents
        bb (o3d.geometry.OrientedBoundingBox): The boundingbox that needs to be cropped from the pointcloud
        samplesize (Optional float): When the results needs to be downsampled.
            
    Returns:
        PointCloudNode containing the pointcloud of the cropped region
    """
    myCroppedPCD=pointcloud.resource.crop(bb)
    if len(np.asarray(myCroppedPCD.points)) > 0:
        if sn.resolution and len(np.asarray(myCroppedPCD.points)) >= 1000:
            myCroppedPCD = myCroppedPCD.voxel_down_sample(sn.resolution)
        
        try:
            myCroppedPCDNode=PointCloudNode(resource=myCroppedPCD,
                                        name = element.get_name().split("-")[0] + "-CROPPEDPCD",
                                        subject=element.get_name().split("-")[0] + "-CROPPEDPCD",
                                        isDerivedFromGeometry=pointcloud.subject,
                                        processed = False)
            myCroppedPCDNode.save_resource(sn.projectProcessingPath)
            return myCroppedPCDNode
        
        except:
            return None
    else:
        return None

def add_to_croppedpcd(pointcloud: PointCloudNode, element: BIMNode, target: PointCloudNode ,bb: o3d.geometry.OrientedBoundingBox, sn: SessionNode) -> PointCloudNode:
    """adds a part to a point cloud.

    Args:
        
            
    Returns:
        PointCloudNode containing the pointcloud of the cropped region
    """
    myCroppedPCD=pointcloud.resource.crop(bb)
    if len(np.asarray(myCroppedPCD.points)) > 0:
        if sn.resolution and len(np.asarray(myCroppedPCD.points)) >= 1000:
            myCroppedPCD = myCroppedPCD.voxel_down_sample(sn.resolution)
        target.resource.__iadd__(myCroppedPCD)
        target.save_resource(sn.projectProcessingPath)

def get_BIMNode(subject, sn : SessionNode):
    """Gets the BIMNode  from which it is derived.

    Args:
        subject: the subject from which it was derived
        BIMNodeList (List[BIMNodes]): list of all the BIMNodes in the project
            
    Returns:
        PointCloudNode containing the pointcloud of the cropped region
    """
    for n in sn.bimNodes:
        if n.subject is subject:
            return n

def get_cropped_pcdNode(subject, sn : SessionNode):
    for n in sn.meshpcdNodes:
        if n.subject.split("-")[0] in subject:
            return n

def get_referenceCloudNode(subject, sn : SessionNode):
    for n in sn.meshpcdNodes:
        if n.subject.split("-")[0] in subject:
            return n

def distance_filtering(target: PointCloudNode, reference: PointCloudNode, sn : SessionNode):
    
    d = target.resource.compute_point_cloud_distance(reference.resource)
    
    i = 0
    indeces = []

    while i < len(d):
        if d[i] < sn.Td:
            indeces.append(i)
        i += 1
    
    BIMNode = get_BIMNode(reference.isDerivedFromGeometry, sn=sn)
    
    if len(indeces) > 50:
        filteredpcdNode = PointCloudNode(resource = target.resource.select_by_index(indeces),
                                        subject=BIMNode.get_name().split("-")[0] + "-FILTEREDPCD",
                                        name = BIMNode.get_name().split("-")[0] + "-FILTEREDPCD",
                                        isDerivedFromGeometry=target.subject,
                                        processingType = 'DISTANCE',
                                        processed = False)
        filteredpcdNode.path = os.path.join(sn.projectProcessingPath, filteredpcdNode.get_name() + ".pcd")
        if os.path.exists(filteredpcdNode.path):
            print("Path already exisits")
            filteredpcdNode.path = os.path.join(sn.projectProcessingPath, filteredpcdNode.get_name() + time.strftime("%Y%m%d%H%M%S") +".pcd")
        filteredpcdNode.save_resource(sn.projectProcessingPath)
       
        return filteredpcdNode
    else:
        return None

def normal_filtering(target: PointCloudNode, reference: PointCloudNode, sn : SessionNode ):
    if not target.resource.has_normals():
        target.resource.estimate_normals()
    if not reference.resource.has_normals():
        reference.resource.estimate_normals()

    d = target.resource.compute_point_cloud_distance(reference.resource)

    i = 0
    indeces_d = []

    while i < len(d):
        if d[i] < sn.Td:
            indeces_d.append(i)
        i += 1
    
    BIMNode = get_BIMNode(reference.isDerivedFromGeometry, sn=sn)
    
    if len(indeces_d) > 50:
        indeces_n = []
        kdtree = o3d.geometry.KDTreeFlann(reference.resource)
        for id in indeces_d:
            [k, idx, d] = kdtree.search_radius_vector_3d(target.resource.points[id], 0.1)
            matched = False
            i = 0
            while not matched and i < len(idx) and len(idx) > 0:
                if np.abs(np.dot(np.asarray(target.resource.normals[id]), np.asarray(reference.resource.normals[idx[i]]))) > sn.Tdot:
                    indeces_n.append(id)
                    matched = True
                i += 1
        if len(indeces_n) > 50:
            filteredpcdNode = PointCloudNode(resource = target.resource.select_by_index(indeces_n),
                                        subject=BIMNode.get_name().split("-")[0] + "-FILTEREDPCD",
                                        name = BIMNode.get_name().split("-")[0] + "-FILTEREDPCD",
                                        isDerivedFromGeometry=target.subject,
                                        processingType = 'NORMALS',
                                        processed = False)
            filteredpcdNode.path = os.path.join(sn.projectProcessingPath, filteredpcdNode.get_name() + ".pcd")
            if os.path.exists(filteredpcdNode.path):
                print("Path already exisits")
                filteredpcdNode.path = os.path.join(sn.projectProcessingPath, filteredpcdNode.get_name() + time.strftime("%Y%m%d%H%M%S") +".pcd")
            filteredpcdNode.save_resource(sn.projectProcessingPath)
        
            return filteredpcdNode
        else: 
            return None
    else:
        return None

def color_by_LOA(target: PointCloudNode, distances: o3d.utility.DoubleVector, sn : SessionNode):
    i=0
    target.resource.paint_uniform_color([1,1,1])
    # print(len(np.asarray(target.resource.points)))
    # print(len(np.asarray(target.resource.colors)))
    # print(len(np.asarray(distances)))
    while i < len(distances):
        distance = distances[i]
        
        if distance < sn.t00:
            np.asarray(target.resource.colors)[i] = [1,0,0]
        if distance < sn.t10:
            np.asarray(target.resource.colors)[i] = [1,0.76,0]
        if distance < sn.t20:
            np.asarray(target.resource.colors)[i] = [1,1,0]
        if distance < sn.t30:
            np.asarray(target.resource.colors)[i] = [0,1,0]
        i += 1
    
    pcdResultDirectory = os.path.join(sn.resultsPath, "PCD")
    if not os.path.isdir(pcdResultDirectory): #check if the folder exists
        os.mkdir(pcdResultDirectory)

    target.save_resource(pcdResultDirectory)

def compute_LOA(target: PointCloudNode, reference: PointCloudNode,sn : SessionNode):
    d = target.resource.compute_point_cloud_distance(reference.resource)
    
    LOA10Inliers = 0
    LOA20Inliers = 0
    LOA30Inliers = 0
    LOA00Inliers = 0
    totalPoints = len(d)

    for dis in np.asarray(d):
        if dis < sn.t00:
            LOA00Inliers += 1
        if dis < sn.t10:
            LOA10Inliers += 1
        if dis < sn.t20:
            LOA20Inliers += 1
        if dis < sn.t30: 
            LOA30Inliers += 1
    if LOA00Inliers < totalPoints:
        prct = LOA00Inliers/totalPoints *100
    
    if LOA00Inliers > 0:
        LOA10 = LOA10Inliers/LOA00Inliers
        LOA20 = LOA20Inliers/LOA00Inliers
        LOA30 = LOA30Inliers/LOA00Inliers

        color_by_LOA(target = target, distances=d, sn = sn)

        return [LOA10, LOA20, LOA30]
    else:
        return None

def LOA_to_csv(element : BIMNode, sn : SessionNode):
    data = [element.name, element.globalId, element.key, element.accuracy, element.LOA10, element.LOA20, element.LOA30]
    sn.csvWriter.writerow(data)

def LOA_to_xlsx(element : BIMNode, sn : SessionNode):

    sn.worksheet.write(sn.xlsxRow, 0, element.name)
    sn.worksheet.write(sn.xlsxRow, 1, element.globalId)
    sn.worksheet.write(sn.xlsxRow, 2, element.key)
    sn.worksheet.write(sn.xlsxRow, 3, element.accuracy)
    sn.worksheet.write(sn.xlsxRow, 4, element.LOA10)
    sn.worksheet.write(sn.xlsxRow, 5, element.LOA20)
    sn.worksheet.write(sn.xlsxRow, 6, element.LOA30)

    sn.xlsxRow +=1


def LOA_to_mesh(element : BIMNode, sn : SessionNode):
    
    meshResultDirectory = os.path.join(sn.resultsPath, "PLY")
    if not os.path.isdir(meshResultDirectory): #check if the folder exists
        os.mkdir(meshResultDirectory)
    
    meshResultName = element.name
    meshResultFileName = meshResultName + ".ply"
    meshResultPath = os.path.join(meshResultDirectory, meshResultFileName)

    if not element.accuracy == "NO LOA":
        if element.resource or os.path.exists(element.path):
            loaded = False
            if not element.resource and element.path:
                #Load the meshobject from disk
                element.resource = o3d.io.read_triangle_mesh(element.path)
                loaded = True
        result = element.resource
        if element.accuracy == "LOA30":
            result.paint_uniform_color([0,1,0])
        elif element.accuracy == "LOA20":
            result.paint_uniform_color([1,1,0])
        elif element.accuracy == "LOA10":
            result.paint_uniform_color([1,0.76,0])
        elif element.accuracy == "LOA00":
            result.paint_uniform_color([1,0,0])
        else:
            result.paint_uniform_color([1,1,1])
        
        if loaded:
            element.resource = None
        
        o3d.io.write_triangle_mesh(meshResultPath,result)
        # print("OBJ file saved in %s" %meshResultPath)

    else: 
        print("ERROR no LOAs where computed in previous steps") 

def optimize_camera(node: Node):
    
    fov = np.pi / 3 #60 #degrees
    
    #determine extrinsic camera parameters
    extrinsic = np.empty((1,3), dtype=float)
    
    if node.resource is not None:
        box=node.resource.get_oriented_bounding_box()
        c = box.get_center()
        u= box.extent[0]

        d_w=math.cos(fov/2)*u
        #determine c_i
        rotation_matrix=box.R
        pcd = o3d.geometry.PointCloud()
        array=np.array([[c[0],c[1],c[2]+d_w]])
        pcd.points = o3d.utility.Vector3dVector(array)
        pcd.rotate(rotation_matrix, center =c)
        c_i=np.asarray(pcd.points[0])

        up = [0, 0, 1]  # camera orientation

    return c,c_i,up

def create_detail_image(resultpcd : PointCloudNode, sn : SessionNode, width = 640, height = 480):

    referenceNode = get_referenceCloudNode(subject = resultpcd.subject, sn = sn)
    bimnode = get_BIMNode(subject = referenceNode.isDerivedFromGeometry, sn = sn)
    
    #generate scene
    render = o3d.visualization.rendering.OffscreenRenderer(width,height)

    #Determine the materials for visualization 
    mtl=o3d.visualization.rendering.MaterialRecord()
    mtl.base_color = [1.0, 1.0, 1.0, 1.0]  # RGBA
    mtl.shader = "defaultUnlit"
    
    mtlline=o3d.visualization.rendering.MaterialRecord()
    mtlline.base_color = [0.0, 0.0, 0.0, 0.0]  # RGBA
    mtlline.shader = "defaultUnlit"
    
    #set camera
    # # Look at the origin from the front (along the -Z direction, into the screen), with Y as Up.
    center, eye, up = optimize_camera(bimnode)
    render.scene.camera.look_at(center, eye, up)
    
    #add geometries
    if not bimnode.resource:
        bimnode.get_resource()
    #Add the BIM elements wireframe to the scene to have a bether understanding of the object
    wireframe = o3d.geometry.LineSet.create_from_triangle_mesh(bimnode.resource)
    wireframe.paint_uniform_color([0,0,0])
    
    render.scene.add_geometry("test",resultpcd.resource,mtl)
    render.scene.add_geometry("lines", wireframe, mtlline)

    #render the image
    img = render.render_to_image()
    imageResultDirectory = os.path.join(sn.resultsPath, "IMAGES")
    if not os.path.isdir(imageResultDirectory):
        os.mkdir(imageResultDirectory)
    detailImageResultName = bimnode.name
    detailImageResultFileName = detailImageResultName + ".png"
    detailImageResultPath = os.path.join(imageResultDirectory, detailImageResultFileName)

    o3d.io.write_image(detailImageResultPath, img)
    bimnode.detailImagePath = detailImageResultPath

def create_report(resultpcd : PointCloudNode, sn : SessionNode):

    referenceNode = get_referenceCloudNode(subject = resultpcd.subject, sn = sn)
    bimnode = get_BIMNode(subject = referenceNode.isDerivedFromGeometry, sn = sn)

    #Load both the detail and overview image of the object
    detail = Image.open(bimnode.detailImagePath)
    overview = Image.open(bimnode.overviewImagePath)

    info = Image.new('RGB', (int(overview.width/2), int(overview.height/2)), color = (255,255,255))
    d = ImageDraw.Draw(info)
    d.text((10,10),bimnode.name, fill=(0,0,0))
    d.text((10,20), time.strftime("%Y%m%d-%H%M%S"), fill=(0,0,0))
    
    d.text((10,70),"LOA10  (%sm - %sm)" %(sn.t20,sn.t10), fill=(0,0,0))
    if np.round(bimnode.LOA10, decimals = 2) < sn.p10:
        d.text((200,70), str(np.round(bimnode.LOA10*100, decimals = 2)) + "%", fill=(255,0,0))
    else:
        d.text((200,70), str(np.round(bimnode.LOA10*100, decimals = 2))+ "%", fill=(0,255,0))
    
    d.text((10,80), "LOA20  (%sm - %sm)" %(sn.t30, sn.t20), fill=(0,0,0))
    if np.round(bimnode.LOA20, decimals = 2) < sn.p20:
        d.text((200,80), str(np.round(bimnode.LOA20*100, decimals = 2))+ "%", fill=(255,0,0))
    else:
        d.text((200,80), str(np.round(bimnode.LOA20*100, decimals = 2))+ "%", fill=(0,255,0))

    d.text((10,90), "LOA30  (%sm - %sm)" %(0, sn.t30), fill=(0,0,0))
    if np.round(bimnode.LOA30, decimals = 2) < sn.p30:
        d.text((200,90), str(np.round(bimnode.LOA30*100, decimals = 2))+ "%", fill=(255,0,0))
    else:
        d.text((200,90), str(np.round(bimnode.LOA30*100, decimals = 2))+ "%", fill=(0,255,0))
    

    resizedInfo= info.resize((overview.width,overview.height), Image.ANTIALIAS)
    info_overview = Image.new('RGB', (overview.width, overview.height*2))
    info_overview.paste(resizedInfo, (0,0))
    info_overview.paste(overview, (0,overview.height))

    newHeigth = 2*overview.height
    hpercent = (newHeigth/float(detail.height))
    wsize = int((float(detail.width)*float(hpercent)))
    resizedImage = detail.resize((wsize, newHeigth), Image.ANTIALIAS)

    report = Image.new('RGB', (overview.width + wsize, overview.height*2))
    report.paste(info_overview,(0,0))
    report.paste(resizedImage, (info_overview.width, 0))

    reportsResultDirectory = os.path.join(sn.resultsPath, "Reports")
    if not os.path.isdir(reportsResultDirectory):
        os.mkdir(reportsResultDirectory)
    reportName = bimnode.name
    reportFileName = reportName + ".png"
    reportPath = os.path.join(reportsResultDirectory, reportFileName)

    report.save(reportPath)


# import resource
import copy
import math
import os
import shutil
import time
import unittest
from multiprocessing.sharedctypes import Value

import cv2
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt
import ifcopenshell
import numpy as np
import open3d as o3d
import PIL
import pye57
import rdflib
from geomapi.nodes import *
from ifcopenshell.util.selector import Selector
from rdflib import RDF, RDFS, Graph, Literal, URIRef


class TestNode(unittest.TestCase):

################################## SETUP/TEARDOWN CLASS ######################
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") 

        #ONTOLOGIES
        cls.exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#')
        cls.geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') #coordinate system information
        cls.gom=rdflib.Namespace('https://w3id.org/gom#') # geometry representations => this is from mathias
        cls.omg=rdflib.Namespace('https://w3id.org/omg#') # geometry relations
        cls.fog=rdflib.Namespace('https://w3id.org/fog#')
        cls.v4d=rdflib.Namespace('https://w3id.org/v4d/core#')
        cls.openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')
        cls.e57=rdflib.Namespace('http://libe57.org#')
        cls.xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        cls.ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')

        #PCDGRAPH
        cls.graphPath1=os.path.join(cls.path,'pcdGraph.ttl')
        cls.graph1=Graph().parse(cls.graphPath1)

        #MESHGRAPH
        cls.graphPath2=os.path.join(cls.path,'meshGraph.ttl')
        cls.graph2=Graph().parse(cls.graphPath2)
                
        #PCD1 
        cls.path1=os.path.join(cls.path,'PCD',"academiestraat week 22 a 20.pcd")  
        # cls.pcd1= o3d.io.read_point_cloud(cls.path1)
        cls.subject1=next(s for s in cls.graph1.subjects(RDF.type) if "file:///academiestraat_week_22_a_20" in s.toPython())

        #MESH2 
        cls.path2=os.path.join(cls.path,'MESH',"Precast Stair_Stair_1181900 Landing 1.obj")  
        cls.mesh2 = o3d.io.read_triangle_mesh(cls.path2)
        cls.box2=cls.mesh2.get_oriented_bounding_box()
        cls.cartesianTransform2=gt.get_cartesian_transform(translation=cls.box2.get_center())
        cls.subject2=next(s for s in cls.graph2.subjects(RDF.type) if "file:///Precast_Stair_Stair_1181900_Landing_1" in s.toPython())

        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)

        #TIME TRACKING           
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')

    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        if os.path.exists(cls.resourcePath):
            shutil.rmtree(cls.resourcePath)

################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))
################################## TEST FUNCTIONS ######################

    def test_Node_creation_from_subject(self):
        #subject
        subject='myNode'
        node= GeometryNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'file:///'+subject)

    def test_Node_creation_with_cartesianBounds(self):
        #None
        cartesianBounds=None
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertIsNone(node.cartesianBounds)

        #np.array(6x1)
        cartesianBounds=np.array([1,2,3,4,5,6])
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.cartesianBounds.size,6)   

        #list
        cartesianBounds=[1,2,3,4,5,6]
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.cartesianBounds.size,6)   

        #np.array(other)
        cartesianBounds=np.array([[1,2],[3,4],[5,6]])
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.cartesianBounds.size,6)   

        #orientedBoundingBox
        node= GeometryNode(cartesianBounds=self.box2)
        self.assertEqual(node.cartesianBounds.size,6)   

        #Vector3dVector
        cartesianBounds=gt.get_oriented_bounds(np.array([1,2,3,4,5,6]) )
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.cartesianBounds.size,6)   

        #orientedBounds
        cartesianBounds=np.asarray(gt.get_oriented_bounds(np.array([1,2,3,4,5,6]) ))
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.cartesianBounds.size,6)   

        #error
        self.assertRaises(ValueError,GeometryNode,cartesianBounds='thisisnotacartesianbound')


    def test_Node_creation_with_orientedBounds(self):
        #None
        orientedBounds=None
        node= GeometryNode(orientedBounds=orientedBounds)
        self.assertIsNone(node.orientedBounds)

        #np.array(8x3)
        orientedBounds=np.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12],[13,14,15],[16,17,18],[19,20,21],[22,23,24]])
        node= GeometryNode(orientedBounds=orientedBounds)
        self.assertEqual(node.orientedBounds.size,24)   

        #list
        orientedBounds=[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24]
        node= GeometryNode(orientedBounds=orientedBounds)
        self.assertEqual(node.orientedBounds.size,24)   

        #np.array(other)
        orientedBounds=np.array([[1,4,7,10,13,16,19,22],
                                  [2,5,8,11,14,17,20,23],
                                  [3,6,9,12,15,18,21,24]])
        node= GeometryNode(orientedBounds=orientedBounds)
        self.assertEqual(node.orientedBounds.size,24)   

        #orientedBoundingBox
        node= GeometryNode(orientedBounds=self.box2)
        self.assertEqual(node.orientedBounds.size,24)   

        #Vector3dVector
        points=o3d.utility.Vector3dVector(np.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12],[13,14,15],[16,17,18],[19,20,21],[22,23,24]]))   
        node= GeometryNode(orientedBounds=points)
        self.assertEqual(node.orientedBounds.size,24)   

        #error
        self.assertRaises(ValueError,GeometryNode,orientedBounds='thisisnotaorientedbound')


    def test_Node_creation_with_orientedBoundingBox(self):
        #None
        orientedBoundingBox=None
        node= GeometryNode(orientedBoundingBox=orientedBoundingBox)
        self.assertIsNone(node.orientedBoundingBox)

        #box
        node= GeometryNode(orientedBoundingBox=self.box2)
        self.assertAlmostEqual(node.orientedBoundingBox.get_min_bound()[0],self.box2.get_min_bound()[0],delta=0.01)   

        #np.array(nx3)
        orientedBoundingBox=self.mesh2.get_oriented_bounding_box()
        points=np.asarray(orientedBoundingBox.get_box_points())
        node= GeometryNode(orientedBoundingBox=points)
        self.assertAlmostEqual(node.orientedBoundingBox.get_min_bound()[0],orientedBoundingBox.get_min_bound()[0],delta=0.01)   

        #geometry
        node= GeometryNode(orientedBoundingBox=self.mesh2)
        self.assertAlmostEqual(node.orientedBoundingBox.get_min_bound()[0],orientedBoundingBox.get_min_bound()[0],delta=0.01)   

        #Vector3dVector
        orientedBoundingBox=self.mesh2.get_oriented_bounding_box()
        points=orientedBoundingBox.get_box_points()
        node= GeometryNode(orientedBoundingBox=points)
        self.assertAlmostEqual(node.orientedBoundingBox.get_min_bound()[0],orientedBoundingBox.get_min_bound()[0],delta=0.01)   

    def test_Node_creation_with_cartesianTransform(self):
        #None
        cartesianTransform=None
        node= GeometryNode(cartesianTransform=cartesianTransform)
        self.assertIsNone(node.cartesianTransform)

        #center np.array(1x3)        
        center=self.box2.get_center()
        node= GeometryNode(cartesianTransform=center)
        self.assertAlmostEqual(node.cartesianTransform[0,3],center[0],delta=0.01)   

        #cartesianBounds np.array(1x6)
        cartesianBounds=gt.get_cartesian_bounds(self.box2)
        node= GeometryNode(cartesianTransform=cartesianBounds)
        self.assertAlmostEqual(node.cartesianTransform[0,3],(cartesianBounds[0]+cartesianBounds[1])/2,delta=0.01)   

        #orientedBounds np.array(8x3)
        orientedBounds=np.asarray(self.box2.get_box_points())
        node= GeometryNode(cartesianTransform=orientedBounds)
        self.assertAlmostEqual(node.cartesianTransform[0,3],center[0],delta=0.01)   

        #np.ndarray(nx3)
        array=np.concatenate((orientedBounds,orientedBounds),0)
        node= GeometryNode(cartesianTransform=array)
        self.assertAlmostEqual(node.cartesianTransform[0,3],center[0],delta=0.01)   

        #geometry
        node= GeometryNode(cartesianTransform=self.mesh2)
        self.assertAlmostEqual(node.cartesianTransform[0,3],self.mesh2.get_center()[0],delta=0.01)   

        #Vector3dVector
        points=self.box2.get_box_points()
        node= GeometryNode(cartesianTransform=points)
        self.assertAlmostEqual(node.cartesianTransform[0,3],center[0],delta=0.01)   

    def test_get_oriented_bounding_box(self):

        #empty
        node=GeometryNode()        
        self.assertIsNone(node.get_oriented_bounding_box())

        #orientedBoundingBox
        node= GeometryNode(orientedBoundingBox=self.box2)
        self.assertAlmostEqual(node.get_oriented_bounding_box().get_min_bound()[0],self.box2.get_min_bound()[0],delta=0.01)   

        #cartesianBounds
        cartesianBounds=gt.get_cartesian_bounds(self.mesh2)
        node= GeometryNode(cartesianBounds=cartesianBounds)
        self.assertAlmostEqual(node.get_oriented_bounding_box().get_min_bound()[0],self.mesh2.get_axis_aligned_bounding_box().get_min_bound()[0],delta=0.01)   

        #orientedBounds
        orientedBounds=gt.get_oriented_bounds(cartesianBounds)
        node= GeometryNode(orientedBounds=orientedBounds)
        self.assertAlmostEqual(node.get_oriented_bounding_box().get_min_bound()[0],self.mesh2.get_axis_aligned_bounding_box().get_min_bound()[0],delta=0.01)   

    def test_get_cartesian_transform(self):

        #emtpy
        node=GeometryNode()        
        self.assertIsNone(node.get_cartesian_transform())

        #cartesianTransform
        node=GeometryNode(cartesianTransform=self.cartesianTransform2)
        self.assertEqual(node.get_cartesian_transform()[0,3], self.cartesianTransform2[0,3])

        #cartesianBounds
        cartesianBounds=gt.get_cartesian_bounds(self.mesh2)
        node=GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.get_cartesian_transform()[0,3], (cartesianBounds[0]+cartesianBounds[1])/2)

        #orientedBounds
        orientedBounds=np.asarray(self.box2.get_box_points())
        node=GeometryNode(orientedBounds=orientedBounds)
        self.assertEqual(node.get_cartesian_transform()[0,3], np.mean(orientedBounds,0)[0])

        #orientedBoundingBox
        node=GeometryNode(orientedBoundingBox=self.box2)
        self.assertEqual(node.get_cartesian_transform()[0,3], np.mean(orientedBounds,0)[0])

        #resource
        node=GeometryNode(resource=self.mesh2)
        self.assertEqual(node.get_cartesian_transform()[0,3],  self.mesh2.get_center()[0])

    def test_get_cartesian_bounds(self):
        #emtpy
        node=GeometryNode()        
        self.assertIsNone(node.get_cartesian_bounds())

        #cartesianBounds
        cartesianBounds=gt.get_cartesian_bounds(self.mesh2)
        node=GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.get_cartesian_bounds()[0], cartesianBounds[0])

        #orientedBounds
        orientedBounds=np.asarray(self.box2.get_box_points())
        node=GeometryNode(orientedBounds=orientedBounds)
        self.assertAlmostEqual(node.get_cartesian_bounds()[0], self.box2.get_min_bound()[0],delta=0.01)

        #orientedBoundingBox
        node=GeometryNode(orientedBoundingBox=self.box2)
        self.assertEqual(node.get_cartesian_bounds()[0],  self.box2.get_min_bound()[0])

        #resource
        node=GeometryNode(resource=self.mesh2)
        self.assertEqual(node.get_cartesian_bounds()[0],  self.mesh2.get_min_bound()[0])


    def test_get_oriented_bounds(self):
        #emtpy
        node=GeometryNode()        
        self.assertIsNone(node.get_oriented_bounds())

        #cartesianBounds
        cartesianBounds=gt.get_cartesian_bounds(self.mesh2)
        node=GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.get_oriented_bounds()[0,0], cartesianBounds[0])

        #orientedBounds
        orientedBounds=np.asarray(self.box2.get_box_points())
        node=GeometryNode(orientedBounds=orientedBounds)
        self.assertEqual(node.get_oriented_bounds()[2,2], orientedBounds[2,2])

        #orientedBoundingBox
        node=GeometryNode(orientedBoundingBox=self.box2)
        self.assertEqual(node.get_oriented_bounds()[2,2],  np.asarray(self.box2.get_box_points())[2,2])

        #resource
        node=GeometryNode(resource=self.mesh2)
        self.assertEqual(node.get_oriented_bounds()[2,2],  np.asarray(self.box2.get_box_points())[2,2])

    def test_get_center(self):
        #emtpy
        node=GeometryNode()        
        self.assertIsNone(node.get_center())

        #cartesianBounds
        cartesianBounds=gt.get_cartesian_bounds(self.mesh2)
        node=GeometryNode(cartesianBounds=cartesianBounds)
        self.assertEqual(node.get_center()[0], (cartesianBounds[0]+cartesianBounds[1])/2)

        #orientedBounds
        orientedBounds=np.asarray(self.box2.get_box_points())
        node=GeometryNode(orientedBounds=orientedBounds)
        self.assertEqual(node.get_center()[0], np.mean(orientedBounds,0)[0])

        #orientedBoundingBox
        node=GeometryNode(orientedBoundingBox=self.box2)
        self.assertEqual(node.get_center()[0],  self.box2.get_center()[0])

        #resource
        node=GeometryNode(resource=self.mesh2)
        self.assertEqual(node.get_center()[0],  self.mesh2.get_center()[0])


if __name__ == '__main__':
    unittest.main()



import copy
import math
import os
import shutil
import time
import unittest

import cv2
import geomapi.tools.linkeddatatools as ld
import geomapi.utils.geometryutils as gt
import ifcopenshell
import numpy as np
import open3d as o3d
import pye57
import pytest
from ifcopenshell.util.selector import Selector
from rdflib import RDF, RDFS, Graph, Literal, URIRef

 ################################## SETUP/TEARDOWN MODULE ######################

# def setUpModule():
#     #execute once before the module 
#     print('-----------------Setup Module----------------------')

# def tearDownModule():
#     #execute once after the module 
#     print('-----------------TearDown Module----------------------')

class TestLinkeddatatools(unittest.TestCase):

 ################################## SETUP/TEARDOWN CLASS ######################
  
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") #os.pardir, 
       
        #POINTCLOUD - E57_1
        cls.e57Path1=os.path.join(cls.path,'PCD',"week 22 - Lidar.e57")     

        #POINTCLOUD - E57_2
        cls.e57Path2=os.path.join(cls.path,'PCD',"week22_photogrammetry_densecloud - Cloud.e57")  

        #POINTCLOUD - XML
        cls.e57XmlPath=os.path.join(cls.path,'PCD',"week 22 - Lidar.xml") 

        #MESH
        cls.meshPath=os.path.join(cls.path,'MESH',"week22.obj")  
        # cls.mesh= o3d.io.read_triangle_mesh(cls.meshPath)
    
        #IMG1
        cls.imgXmlPath1=os.path.join(cls.path,'IMG',"camera_position.xml")  

        #IMG2
        cls.imgXmlPath2=os.path.join(cls.path,'IMG',"ReferenceLAMBERT08_TAW.xml")  
        
        #IFC1
        cls.bimGraphPath1=os.path.join(cls.path,'bimGraph1.ttl')
        # cls.bimGraph1=Graph().parse(cls.bimGraphPath1)
        cls.ifcPath1=os.path.join(cls.path,'IFC',"Academiestraat_building_1.ifc")  
        # cls.ifc1=ifcopenshell.open(cls.ifcPath1)   
        
        #IFC2 
        cls.bimGraphPath2=os.path.join(cls.path,'bimGraph2.ttl')
        # cls.bimGraph2=Graph().parse(cls.bimGraphPath2)
        cls.ifcPath2=os.path.join(cls.path,'IFC',"Mariakerke_AWV_Conform_3D_BT_l72.ifc")  
        # cls.ifc2=ifcopenshell.open(cls.ifcPath2)  

        #IFC3 
        cls.bimGraphPath3=os.path.join(cls.path,'bimGraph3.ttl')
        # cls.bimGraph3=Graph().parse(cls.bimGraphPath3)
        cls.ifcPath3=os.path.join(cls.path,'IFC',"Academiestraat_parking.ifc")  
        # cls.ifc3=ifcopenshell.open(cls.ifcPath3)   
        
        #IFC4 (IfcDoor)
        cls.bimGraphPath4=os.path.join(cls.path,'bimGraph4.ttl')
        # cls.bimGraph4=Graph().parse(cls.bimGraphPath4)
        cls.ifcPath4=os.path.join(cls.path,'IFC',"B1_ALG_Model.ifc")  
        # cls.ifc4=ifcopenshell.open(cls.ifcPath4)   
        
        #resourceGraph (data)
        cls.resourceGraphPath=os.path.join(cls.path,'resourceGraph.ttl')
        # cls.resourceGraph=Graph().parse(cls.resourceGraphPath)
        # cls.linkedSubjects=[s for s in cls.resourceGraph.subjects(RDF.type)]

        #BIMGRAPH
        cls.bimGraphPath1=os.path.join(cls.path,'bimGraph1.ttl')

        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)
   
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')

    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        if os.path.exists(cls.resourcePath):
            shutil.rmtree(cls.resourcePath)  
        if os.path.exists(os.path.join(os.getcwd(),'resources') ):
            shutil.rmtree(os.path.join(os.getcwd(),'resources') )

################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))

################################## TEST FUNCTIONS ######################

    def test_e57xml_to_nodes(self):
        # e57XML
        nodes=ld.e57xml_to_nodes(e57XmlPath=self.e57XmlPath)
        self.assertEqual(len(nodes),2)
        #0
        self.assertAlmostEqual(nodes[0].cartesianTransform[0,3],0.379,delta=0.01)
        self.assertLess(nodes[0].cartesianTransform[0,0],0)
        self.assertAlmostEqual(nodes[0].cartesianBounds[0],-4.351,delta=0.01)
        #1
        self.assertEqual(nodes[1].cartesianTransform[0,3],0.0)
        self.assertEqual(nodes[1].cartesianTransform[0,0],-1)
        self.assertAlmostEqual(nodes[1].cartesianBounds[0],-5.69,delta=0.01)
        
        #with kwargs
        nodes=ld.e57xml_to_nodes(e57XmlPath=self.e57XmlPath,myattrib=1)
        self.assertEqual(nodes[0].myattrib,1)

        # with getResource 
        # nodes=ld.e57xml_to_nodes(e57XmlPath=self.e57XmlPath,getResource=True)
        # geometries=[n.resource for n in nodes]
        # self.assertEqual(len(geometries),2)

        # error when invalid extension
        self.assertRaises(ValueError,ld.e57xml_to_nodes,e57XmlPath=os.path.join(self.path,'PCD',"myXML.pcd") )

        # error when path doesn't exist
        self.assertRaises(ValueError,ld.e57xml_to_nodes,e57XmlPath=os.path.join(self.path,'PCD',"myXML.xml")    )


    def test_img_xml_to_nodes(self):

        #XML1
        nodes=ld.img_xml_to_nodes(self.imgXmlPath1) 
        self.assertEqual(len(nodes),119)
        self.assertIsNone(nodes[10].resource)
        #0
        self.assertAlmostEqual(nodes[0].cartesianTransform[0,3],6.645,delta=0.01)
        #1
        self.assertAlmostEqual(nodes[1].cartesianTransform[0,3],6.646,delta=0.01)

        #kwargs
        nodes=ld.img_xml_to_nodes(self.imgXmlPath1,sensor='DJI - Phantom 4') 
        self.assertEqual(nodes[5].sensor,'DJI - Phantom 4')
        
        #getResource
        # nodes=ld.img_xml_to_nodes(self.imgXmlPath1,getResource=True) 
        # geometries=[n.resource for n in nodes]
        # self.assertEqual(len(geometries),119)

        #XML2
        nodes=ld.img_xml_to_nodes(self.imgXmlPath2) 
        self.assertEqual(len(nodes),21)
        self.assertAlmostEqual(nodes[0].cartesianTransform[0,3],600575.992,delta=0.01)

        # error when invalid extension
        self.assertRaises(ValueError,ld.img_xml_to_nodes,xmlPath=os.path.join(self.path,'PCD',"myXML.e57") )

        # error when path doesn't exist
        self.assertRaises(ValueError,ld.img_xml_to_nodes,xmlPath=os.path.join(self.path,'IMG',"myXML.xml")    )

    def test_e57header_to_nodes(self):

        #E571
        nodes=ld.e57header_to_nodes(e57Path=self.e57Path1)
        self.assertEqual(len(nodes),2)
        self.assertIsNone(nodes[0].resource)
        #0
        self.assertAlmostEqual(nodes[0].cartesianTransform[0,3],0.510,delta=0.01)
        self.assertLess(nodes[0].cartesianTransform[0,0],0)
        self.assertAlmostEqual(nodes[0].cartesianBounds[0],-36.77,delta=0.01)
        #1
        self.assertAlmostEqual(nodes[1].cartesianTransform[0,3],4.524,delta=0.01)
        self.assertAlmostEqual(nodes[1].cartesianTransform[0,0],0.941,delta=0.01)
        self.assertAlmostEqual(nodes[1].cartesianBounds[0],-28.360,delta=0.01)
        
        #with kwargs
        nodes=ld.e57xml_to_nodes(e57XmlPath=self.e57XmlPath,myattrib=1)
        self.assertEqual(nodes[0].myattrib,1)

        # # with getResource 
        # nodes=ld.e57xml_to_nodes(e57Path=self.e57XmlPath,getResource=True)
        # geometries=[n.resource for n in nodes]
        # self.assertEqual(len(geometries),2)

        #E572
        ld.e57header_to_nodes(e57Path=self.e57Path2)
        self.assertAlmostEqual(nodes[0].cartesianTransform[0,3],0.379,delta=0.01)
        self.assertLess(nodes[0].cartesianTransform[0,0],0)
        self.assertAlmostEqual(nodes[0].cartesianBounds[0],-4.351,delta=0.01)

        # error when invalid extension
        self.assertRaises(ValueError,ld.e57header_to_nodes,e57Path=os.path.join(self.path,'PCD',"myXML.xml") )

        # error when path doesn't exist
        self.assertRaises(ValueError,ld.e57header_to_nodes,e57Path=os.path.join(self.path,'PCD',"myXML.e57")    )

    def test_ifc_to_nodes_errors(self):
        # error when invalid extension
        self.assertRaises(ValueError,ld.ifc_to_nodes,ifcPath=os.path.join(self.path,'IFC',"myXML.xml") )

        # error when path doesn't exist
        self.assertRaises(ValueError,ld.ifc_to_nodes,ifcPath=os.path.join(self.path,'IFC',"myXML.ifc")    )


    def test_ifc_to_nodes(self):
        #IFC1
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath1)
        self.assertEqual(len(nodes), 1192)
        #IFC2
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath2)
        self.assertEqual(len(nodes), 66)
        #IFC3
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath3)
        self.assertEqual(len(nodes), 2705)
        #IFC4
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath4)
        self.assertEqual(len(nodes), 17922)

    def test_ifc_to_nodes_walls(self):
        #IFC1
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath1, classes='.IfcWallStandardCase',getResource=True)
        self.assertEqual(len(nodes), 309)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries),309)
        
        #IFC2
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath2, classes='.IfcWallStandardCase',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(nodes), 0)
        #IFC3
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath3, classes='.IfcWallStandardCase',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 340)
        #IFC4
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath4, classes='.IfcWallStandardCase',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 1603)

    def test_ifc_to_nodes_slabs(self):
        #IFC1
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath1, classes='.IfcSlab',getResource=True)
        self.assertEqual(len(nodes), 237)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries),237)
        #IFC2
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath2, classes='.IfcSlab',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 0)
        #IFC3
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath3, classes='.IfcSlab',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 1150)
        #IFC4
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath4, classes='.IfcSlab',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 415)
    
    def test_ifc_to_nodes_windows_and_doors(self):
      
        #IFC2
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath2, classes='.IfcWindow | .ifcDoor',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 0)
        #IFC3
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath3, classes='.IfcWindow | .ifcDoor',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 0)
        #IFC4
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath4, classes='.IfcWindow | .ifcDoor',getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(geometries), 863)

    def test_nodes_to_graph(self):
        nodes=ld.ifc_to_nodes(ifcPath=self.ifcPath1)
        graph=ld.nodes_to_graph(nodes)
        subjects=[s for s in graph.subjects(RDF.type)]
        self.assertEqual(len(nodes), len(subjects))

    def test_graph_to_nodes(self):
        graph=Graph().parse(self.resourceGraphPath)
        nodes=ld.graph_to_nodes(graph=graph)
        subjects=[s for s in graph.subjects(RDF.type)]
        self.assertEqual(len(nodes),len(subjects))

        #with getResource
        nodes=ld.graph_to_nodes(graph=graph,getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(nodes),len(geometries))

    def test_graphpath_to_nodes(self):
        nodes=ld.graph_path_to_nodes(graphPath=self.resourceGraphPath)
        graph=Graph().parse(self.resourceGraphPath)
        subjects=[s for s in graph.subjects(RDF.type)]
        self.assertEqual(len(nodes),len(subjects))

        #with getResource
        nodes=ld.graph_path_to_nodes(graphPath=self.resourceGraphPath,getResource=True)
        geometries=[n.resource for n in nodes]
        self.assertEqual(len(nodes),len(geometries))

    def test_select_k_nearest_nodes(self):
        nodes=ld.graph_path_to_nodes(graphPath=self.bimGraphPath1,getResource=True)
        list,distances=ld.select_k_nearest_nodes(nodes[0],nodes)
        self.assertEqual(len(list),10)

        #error k<=0
        self.assertRaises(ValueError,ld.select_k_nearest_nodes,node=nodes[0],nodelist=nodes,k=-5 )

    def test_select_nodes_with_centers_in_radius(self):
        nodes=ld.graph_path_to_nodes(graphPath=self.bimGraphPath2,getResource=True)
        list,distances=ld.select_nodes_with_centers_in_radius(nodes[0],nodes,r=10)
        self.assertEqual(len(list),2)
        for d in distances:
            self.assertLess(d,10)

        # no nodes
        list,distances=ld.select_nodes_with_centers_in_radius(nodes[0],nodes,r=0.01)
        self.assertIsNone(list)
        self.assertIsNone(distances)

        #error r<=0
        self.assertRaises(ValueError,ld.select_nodes_with_centers_in_radius,r=-5 )

    def test_select_nodes_with_centers_in_bounding_box(self): 
        nodes=ld.graph_path_to_nodes(graphPath=self.bimGraphPath3,getResource=True)
        list=ld.select_nodes_with_centers_in_bounding_box(nodes[0],nodes,u=5,v=5,w=5)
        self.assertEqual(len(list),2)

        # no nodes
        node=nodes[0]
        del nodes[0]
        list=ld.select_nodes_with_centers_in_bounding_box(node,nodes,u=0,v=0,w=0)
        self.assertIsNone(list)

    def test_select_nodes_with_intersecting_bounding_box(self): 
        nodes=ld.graph_path_to_nodes(graphPath=self.bimGraphPath4,getResource=True)
        list=ld.select_nodes_with_intersecting_bounding_box(nodes[0],nodes,u=5,v=5,w=5)
        self.assertEqual(len(list),6)

        # no nodes
        node=nodes[0]
        del nodes[0]
        list=ld.select_nodes_with_intersecting_bounding_box(node,nodes,u=-2,v=-2,w=-2)
        self.assertIsNone(list)

    def test_select_nodes_with_intersecting_resources(self):
        nodes=[]
        nodes.extend(ld.graph_path_to_nodes(graphPath=self.resourceGraphPath,getResource=True))
        nodes.extend(ld.graph_path_to_nodes(graphPath=self.bimGraphPath3,getResource=True))

        list= ld.select_nodes_with_intersecting_resources(nodes[3],nodes)
        self.assertEqual(len(list),74)

    def test_get_mesh_representation(self):
        nodes=ld.graph_path_to_nodes(graphPath=self.resourceGraphPath,getResource=True)        
        for n in nodes:
            geometry=ld.get_mesh_representation(n)
            self.assertTrue('TriangleMesh' in str(type(geometry)))


if __name__ == '__main__':
    unittest.main()



"""
Geospatialutils - a Python library for processing coordinate system data.
"""

from xmlrpc.client import Boolean
# from osgeo import gdal #conda install gdal
# from osgeo import osr
# from osgeo import ogr

import xml.etree.ElementTree as ET 
import math
from typing import List
import matplotlib

import ifcopenshell
import ifcopenshell.geom as geom
import ifcopenshell.util
from scipy.spatial.transform import Rotation as R

import trimesh
import fcl

#IMPORT MODULES
import geomapi.utils as ut


def belgian_datum_to_wgs84(latBel:float,lngBel:float) -> tuple:
    """
    Input parameters : Lat, Lng : latitude / longitude in decimal degrees and in Belgian 1972 datum
    Output parameters : LatWGS84, LngWGS84 : latitude / longitude in decimal degrees and in WGS84 datum
    source: http://zoologie.umons.ac.be/tc/algorithms.aspx
    """
    Haut = 0.0   
        
    #conversion to radians
    Lat = (math.pi / 180.0) * latBel
    Lng = (math.pi / 180.0) * lngBel
    
    SinLat = math.sin(Lat)
    SinLng = math.sin(Lng)
    CoSinLat = math.cos(Lat)
    CoSinLng = math.cos(Lng)
    
    dx = -125.8
    dy = 79.9
    dz = -100.5
    da = -251.0
    df = -0.000014192702
    
    LWf = 1.0 / 297.0
    LWa = 6378388
    LWb = (1.0 - LWf) * LWa
    LWe2 = (2.0 * LWf) - (LWf * LWf)
    Adb = 1.0 / (1.0 - LWf)
    
    Rn = LWa / math.sqrt(1.0 - LWe2 * SinLat * SinLat)
    Rm = LWa * (1.0 - LWe2) / (1 - LWe2 * Lat * Lat) ** 1.5
    
    DLat = -dx * SinLat * CoSinLng - dy * SinLat * SinLng + dz * CoSinLat
    DLat = DLat + da * (Rn * LWe2 * SinLat * CoSinLat) / LWa
    DLat = DLat + df * (Rm * Adb + Rn / Adb) * SinLat * CoSinLat
    DLat = DLat / (Rm + Haut)
    
    DLng = (-dx * SinLng + dy * CoSinLng) / ((Rn + Haut) * CoSinLat)
    Dh = dx * CoSinLat * CoSinLng + dy * CoSinLat * SinLng + dz * SinLat
    Dh = Dh - da * LWa / Rn + df * Rn * Lat * Lat / Adb
    
    LatWGS84 = ((Lat + DLat) * 180.0) / math.pi
    LngWGS84 = ((Lng + DLng) * 180.0) / math.pi
    return LatWGS84,LngWGS84

def wgs84_to_belgian_datum(LatWGS84: float,LngWGS84:float) -> tuple:
    """
    Input parameters : Lat, Lng : latitude / longitude in decimal degrees and in WGS84 datum
    Output parameters : LatBel, LngBel : latitude / longitude in decimal degrees and in Belgian datum
    source: http://zoologie.umons.ac.be/tc/algorithms.aspx
    """
    Haut = 0.0 
    
    #conversion to radians
    Lat = (math.pi / 180.0) * LatWGS84
    Lng = (math.pi / 180.0) * LngWGS84
    
    SinLat = math.sin(Lat)
    SinLng = math.sin(Lng)
    CoSinLat = math.cos(Lat)
    CoSinLng = math.cos(Lng)
    
    dx = 125.8
    dy = -79.9
    dz = 100.5
    da = 251.0
    df = 0.000014192702
    
    LWf = 1.0 / 297.0
    LWa = 6378388.0
    LWb = (1.0 - LWf) * LWa
    LWe2 = (2.0 * LWf) - (LWf * LWf)
    Adb = 1.0 / (1.0 - LWf)
    
    Rn = LWa / math.sqrt(1 - LWe2 * SinLat * SinLat)
    Rm = LWa * (1.0 - LWe2) / (1 - LWe2 * Lat * Lat) ** 1.5
    
    DLat = -dx * SinLat * CoSinLng - dy * SinLat * SinLng + dz * CoSinLat
    DLat = DLat + da * (Rn * LWe2 * SinLat * CoSinLat) / LWa
    DLat = DLat + df * (Rm * Adb + Rn / Adb) * SinLat * CoSinLat
    DLat = DLat / (Rm + Haut)
    
    DLng = (-dx * SinLng + dy * CoSinLng) / ((Rn + Haut) * CoSinLat)
    Dh = dx * CoSinLat * CoSinLng + dy * CoSinLat * SinLng + dz * SinLat
    Dh = Dh - da * LWa / Rn + df * Rn * Lat * Lat / Adb
    
    latBel = ((Lat + DLat) * 180.0) / math.pi
    lngBel = ((Lng + DLng) * 180.0) / math.pi
    return  latBel,lngBel

def spherical_coordinates_to_lambert72(latBel :float,lngBel:float) -> tuple:
    """
    Conversion from spherical coordinates to Lambert 72
    Input parameters : lat, lng (spherical coordinates)
    Spherical coordinates are in decimal degrees converted to Belgium datum!
    source: http://zoologie.umons.ac.be/tc/algorithms.aspx
    """
     
    LongRef  = 0.076042943        #'=4°21'24"983
    bLamb  = 6378388 * (1 - (1 / 297))
    aCarre  = 6378388 ** 2
    eCarre  = (aCarre - bLamb ** 2) / aCarre
    KLamb  = 11565915.812935
    nLamb  = 0.7716421928
 
    eLamb = math.sqrt(eCarre)
    eSur2 = eLamb / 2.0
    
    #conversion to radians
    lat = (math.pi / 180.0) * latBel
    lng = (math.pi / 180.0) * lngBel
 
    eSinLatitude  = eLamb * math.sin(lat)
    TanZDemi  = (math.tan((math.pi / 4) - (lat / 2))) * (((1 + (eSinLatitude)) / (1 - (eSinLatitude))) ** (eSur2))
       
 
    RLamb = KLamb * ((TanZDemi) ** nLamb)
 
    Teta = nLamb * (lng - LongRef)
 
    x = 150000 + 0.01256 + RLamb * math.sin(Teta - 0.000142043)
    y = 5400000 + 88.4378 - RLamb * math.cos(Teta - 0.000142043)
    return x,y

def lambert72_to_spherical_coordinates(x :float ,y:float) -> tuple: 
    """"
    Belgian Lambert 1972---> Spherical coordinates
    Input parameters : X, Y = Belgian coordinates in meters
    Output : latitude and longitude in Belgium Datum!
    source: http://zoologie.umons.ac.be/tc/algorithms.aspx
    """
    LongRef = 0.076042943  #      '=4°21'24"983
    nLamb  = 0.7716421928 #
    aCarre = 6378388 ** 2 #
    bLamb = 6378388 * (1 - (1 / 297)) #
    eCarre = (aCarre - bLamb ** 2) / aCarre #
    KLamb = 11565915.812935 #
    
    eLamb = math.sqrt(eCarre)
    eSur2 = eLamb / 2
    
    Tan1  = (x - 150000.01256) / (5400088.4378 - y)
    Lambda = LongRef + (1 / nLamb) * (0.000142043 + math.atan(Tan1))
    RLamb = math.sqrt((x - 150000.01256) ** 2 + (5400088.4378 - y) ** 2)
    
    TanZDemi = (RLamb / KLamb) ** (1 / nLamb)
    Lati1 = 2 * math.atan(TanZDemi)
    
    eSin=0.0
    Mult1=0.0
    Mult2=0.0
    Mult=0.0
    LatiN=0.0
    Diff=1     

    while abs(Diff) > 0.0000000277777:
        eSin = eLamb * math.sin(Lati1)
        Mult1 = 1 - eSin
        Mult2 = 1 + eSin
        Mult = (Mult1 / Mult2) ** (eLamb / 2)
        LatiN = (math.pi / 2) - (2 * (math.atan(TanZDemi * Mult)))
        Diff = LatiN - Lati1
        Lati1 = LatiN
    
    latBel = (LatiN * 180) / math.pi
    lngBel = (Lambda * 180) / math.pi
    return latBel,lngBel

# def wgs84_to_l72(lat:float,long:float) -> tuple:
#     SourceEPSG = 4326  
#     TargetEPSG = 31370

#     source = osr.SpatialReference()
#     source.ImportFromEPSG(SourceEPSG)

#     target = osr.SpatialReference()
#     target.ImportFromEPSG(TargetEPSG)


#     transform = osr.CoordinateTransformation(source, target)
#     point = ogr.Geometry(ogr.wkbPoint)
#     point.SetPoint_2D(0, float(Long), float(Lat))
#     point.Transform(transform)
    
#     x=point.GetX()
#     y=point.GetY()
    
#     return x,y 



#File names
JSON_ID = "SessionData.json"
IMG_EXTENSION = [".jpg", ".png", ".JPG", ".PNG", ".JPEG"]
MESH_EXTENSION = [".obj",".fbx" ]
PCD_EXTENSION = [".pcd", ".pts", ".ply"]
TIME_FORMAT = "%Y-%m-%d %H-%M-%S"
METHODS = ["leastDistance","incremental","raycasting", "fpfh", "super4pcs"]